
class Functive:

    def __init__(self,label,span,position,ids,semiotic) -> None:
        
        self.span = span
        self.label = label  
        self.position = position
        # self.id = semiotic.vocab.encode.get(self.label,None) # this needs ot be fixed by handling UNK!!!
        self.id = ids # Meant to recover ids from HF tokenizer
        self.prob = semiotic.vocab.prob.get(self.label,0)

        # self.parad
        # self.parad_t
        # self.parad_t_soft
        # self.prob_in_parad

        # self.f_type
        # self.type

        # self.head
        # self.children
        # self.subtree
        
    def __repr__(self) -> str:
        return f"Functive({self.label},{self.span})"
    
    def __str__(self) -> str:
        return str(self.label)


from ...util import subsequences
from ...functive import Functive
from math import log

import networkx as nx
from nltk.tokenize import sent_tokenize
        

class Processor:
    """
    Base Processor class
    """

    def __init__(self) -> None:
        pass

    def process(self, sequence: str, semiotic,is_pretokenized=False):
        pass

class disable:
    """
    Disable this step. It returns the input as output #TODO: maybe with the correct type for the pipeline
    """

    def __init__(self) -> None:
        pass

    def process(self, sequence: str, semiotic = None,is_pretokenized=False):
        return sequence #TODO: This should maybe return a list of Functives to preserve the flow of types in the tokenizer pipeline

class SequenceSLG(Processor):
    """
    """

    def __init__(self) -> None:
        self.zipf_factor = .135

    def vocabulary_rank(self,
        voc,
        ):
        # TODO: Zipf factor should (in principle) be computable following Mandelbrot
        voc_rank = {k:(v+1)**self.zipf_factor for v,k in enumerate(voc.keys())}
        return(voc_rank)

    def build_graph_data(
        self,
        string: str,
        voc: dict
        ):
        edge_data = []
        for beginning in range(0, len(string)):
            for end in range(beginning + 1, len(string) + 1):
                subsequence_label = string[beginning:end]
                if subsequence_label not in voc or subsequence_label == string:
                    continue
                edge_data.append(
                    (
                        beginning,
                        end,
                        {
                            "label": subsequence_label,
                            "freq": voc[subsequence_label],
                            "len": len(subsequence_label),
                        },
                    )
                )
        return edge_data

    def chain2seq(self,chain_sp, voc, v_rank = None):
        chain = chain_sp.replace(" ", "")
        lSt = len(chain)
        if v_rank == None:
            voc_rank = self.vocabulary_rank(voc)
        else:
            voc_rank = v_rank
        for c in chain:
            if c not in voc:
                voc[c]=1
                voc_rank[c]=(len(voc)+1)**self.zipf_factor
        graph_data = self.build_graph_data(chain, voc)
        seg_graph_full = nx.DiGraph()
        seg_graph_full.add_edges_from(graph_data)
        # Construct weights
        for edge in seg_graph_full.edges:
            rank = voc_rank[seg_graph_full.edges[edge]["label"]]
            seg_graph_full.edges[edge]["weight"] = rank
        # Find best segmentation out of shortest path
        shortest_path = nx.shortest_path(seg_graph_full, 0, lSt, weight="weight")
        seg_sent = [
            seg_graph_full.edges[edge]["label"] for edge in subsequences(shortest_path, 2)
        ]
        return seg_sent
    
    def process(self, sequence: str, semiotic, is_pretokenized=False):
        
        segments = " ".join(self.chain2seq(sequence, semiotic.vocab.freq)).split()

        spans = []
        for i in range(len(segments)):
            start_i = len("".join(segments[:i]))
            end_i = start_i + len(segments[i])
            spans.append((start_i, end_i))

        tokens = [Functive(segment,span,position,semiotic) for position,(segment,span) in enumerate(zip(segments,spans))]

        chain = sequence.replace(" ", "")
        tree_root = Functive(chain, (0, len(chain)), None, semiotic)
        tree_root.children = tokens

        for token in tokens:
            token.head = tree_root

        return tokens


class TreeSLG(SequenceSLG):
    """
    """
    def __init__(self) -> None:
        super().__init__()

    def vocabulary_rank(self, voc):
        return super().vocabulary_rank(voc)

    def build_graph_data(self, string: str, voc: dict):
        return super().build_graph_data(string, voc)

    def chain2seq(self, chain_sp, voc, v_rank):
        return super().chain2seq(chain_sp, voc, v_rank=v_rank)


    # # Possible binary segmentation functions for binary_segment:
    def bin_seg_f_sum(self, l_units, r_units, voc):
        l_unit = l_units[-1]
        r_unit = r_units[-1]
        return (voc[l_unit] + voc[r_unit])

    def bin_seg_f_logLen(self, l_units, r_units, voc):
        l_unit = l_units[-1]
        r_unit = r_units[-1]
        return (log(voc[l_unit]+len(l_unit)) + log(voc[r_unit])+len(r_unit)) - (abs(len(l_units)-len(r_units)))


    def binary_segment(
        self,
        segmented_chain,
        voc,
        interval = None,
    ):
        chain = "".join(segmented_chain)
        if interval == None:
            interval = (0, len(chain))
        seg_sent_pairs = [
            [segmented_chain[:c], segmented_chain[c:]]
            for c in range(1, len(segmented_chain))
        ]
        cut_scores = []
        for pair in seg_sent_pairs:
            cut_scores.append([self.bin_seg_f_sum(pair[0], pair[1], voc), pair])
        best_cut = sorted(cut_scores, reverse=True)[0][1]
        seg_L = "".join(best_cut[0])
        seg_R = "".join(best_cut[1])
        return (
            (seg_L, (interval[0], interval[0] + len(seg_L))),
            (seg_R, (interval[1] - len(seg_R), interval[1])),
        )


    def chain2tree(self, chain_sp, voc, v_rank = None):
        chain = chain_sp.replace(" ", "")
        if v_rank == None:
            voc_rank = self.vocabulary_rank(voc)
        else:
            voc_rank = v_rank
        stack = [(chain, (0, len(chain)))]
        tree_data = []
        while len(stack) > 0:
            parent = stack[0]
            chain_to_seg = parent[0]
            stack = stack[1:]
            if len(chain_to_seg) == 2:
                l_child = (chain_to_seg[0], (parent[1][0], parent[1][0] + 1))
                r_child = (chain_to_seg[1], (parent[1][1] - 1, parent[1][1]))
            else:
                parent_seg = self.chain2seq(chain_to_seg, voc, voc_rank)
                children = self.binary_segment(parent_seg, interval=parent[1], voc=voc)
                l_child = children[0]
                r_child = children[1]
                for child in [l_child, r_child]:
                    if len(child[0]) > 1:
                        stack.append(child)
            tree_data.extend([(parent, l_child), (parent, r_child)])
        return tree_data
    
    def process(self, sequence: str, semiotic, is_pretokenized=False):

        tree_data = self.chain2tree(sequence,semiotic.vocab.freq)

        seq_tokens = sorted([child for head,child in tree_data if head[0] not in semiotic.vocab.freq and child[0] in semiotic.vocab.freq],key=lambda x:x[-1])
        
        tree_tokens = []
        for head,child in tree_data:
            child_token = Functive(child[0],child[1],None,semiotic)
            if child in seq_tokens:
                child_token.position = seq_tokens.index(child)
            child_token.head = Functive(head[0],head[1],None,semiotic)
            tree_tokens.append(child_token)

        return tree_tokens


class SplitWhitespaces(Processor):
    def __init__(self) -> None:
        super().__init__()
    
    def process(self, sequence: str, semiotic, is_pretokenized):
        segments = sequence.split()

        spans = []
        for i in range(len(segments)):
            start_i = len("".join(segments[:i]))
            end_i = start_i + len(segments[i])
            spans.append((start_i, end_i))

        tokens = [Functive(segment,span,position,semiotic) for position,(segment,span) in enumerate(zip(segments,spans))]

        no_wspace_sequence = "".join(segments)
        tree_root = Functive(no_wspace_sequence, (0, len(no_wspace_sequence)), None, semiotic)
        tree_root.children = tokens

        for token in tokens:
            token.head = tree_root

        return tokens

class SentencesNLTK(Processor):
    """
    NLTK sentence tokenizer. Used for the normalization of initial datasets
    """
    def __init__(self) -> None:
        super().__init__()
    
    def process(self, sequence: str, semiotic=None, is_pretokenized=False):
        
        if is_pretokenized == False:
            return sent_tokenize(sequence)
        else:
            sentences = []
            for pre_token in sequence:
                nltk_sents = sent_tokenize(pre_token)
                for sent in nltk_sents:
                    sentences.append(sent)
            return sentences



# from . import normalizers
from .pre_tokenizers import (
    SequenceSLG,
    PMI_SLG,
)

from tokenizers.normalizers import NFKC, Lowercase, Replace

from tokenizers import (
    normalizers,
    Regex,
)

import string


punctuation = '...—•…–’‘·⁄′¿‐―'
punctuations = (string.punctuation+punctuation).replace("'","").replace('"','').replace('[','').replace(']','')

NormalizeSLG = normalizers.Sequence([
    NFKC(),
    Lowercase(),
    Replace(Regex(f"{[i for i in string.whitespace]}"),""),
    Replace(Regex("["+punctuations+"]"),"-"),
    Replace(Regex("[\'\"\[\]\-]"),""),
    ])

Normalize_w_spacesSLG = normalizers.Sequence([
    NFKC(),
    Lowercase(),
    # Replace(Regex(f"{[i for i in string.whitespace]}"),""),
    Replace(Regex("["+punctuations+"]"),"-"),
    Replace(Regex("[\'\"\[\]\-]"),""),
    ])

Normalize_menoSLG = normalizers.Sequence([
    NFKC(),
    Lowercase(),
    # Replace(Regex(f"{[i for i in string.whitespace]}"),""),
    Replace(Regex("["+punctuations+"]"),"-"),
    Replace(Regex("[\'\"\[\]\-]"),""),
    Replace(Regex("j$"),"i"),
    ])

# TODO: "NormalizeSLG" should be replaced by generic normalizers to be added to a standard HF normalizer sequence
# Below is the model to use for doing it
#
# class CustomNormalizer:
#     def normalize(self, normalized: NormalizedString):
#         # Most of these can be replaced by a `Sequence` combining some provided Normalizer,
#         # (ie Sequence([ NFKC(), Replace(Regex("\s+"), " "), Lowercase() ])
#         # and it should be the prefered way. That being said, here is an example of the kind
#         # of things that can be done here:
#         normalized.nfkc()
#         normalized.filter(lambda char: not char.isnumeric())
#         normalized.replace(Regex("\s+"), " ")
#         normalized.lowercase()
#
# # The custom normalizer should be loaded as follows:
#
# tok.normalizer = Normalizer.custom(CustomNormalizer())


import numpy as np
from scipy.signal import argrelmin, argrelextrema

from ..util import subsequences, plot_scatter_line

class PType:

    def __init__(self,parad,semiotic) -> None:

        self.global_probs = np.array([semiotic.vocab.prob.get(k,0) for k in parad.keys_t])
        if parad.len_truncate == 0:
            self.mass = 0
            self.func_score = 1
        else:
            self.mass = np.mean(self.global_probs)
            self.func_score = self.mass/parad.len_truncate

        self.global_probs_s = np.array([semiotic.vocab.prob.get(k,0) for k in parad.keys_t_soft])
        if parad.len_truncate_soft == 0:
            self.mass_s = 0
            self.func_score_s = 1
        else:
            self.mass_s = np.mean(self.global_probs_s)
            self.func_score_s = self.mass_s/parad.len_truncate_soft



        # self.func = 


class Typer:
    def __init__(self) -> None:
        pass

    def __call__(self,parad_chain):
        types = []
        for parad in parad_chain:
            types.append(PType(parad,parad_chain.semiotic))
        parad_chain.types = types

class TypeChain:

    def __init__(self,parad_chain) -> None:
        self.semiotic = parad_chain.semiotic
        self.len = parad_chain.len
        self.probs = parad_chain.probs
        self.labels = parad_chain.labels
        self.indexes = parad_chain.indexes
        self.types = [ptype for ptype in parad_chain.types]
        self.func_scores = np.array([ptype.func_score_s for ptype in parad_chain.types])
        self.scores_ext = list(argrelextrema(self.func_scores, np.greater)[0])
        self.scores_min = list(argrelmin(self.func_scores)[0])
        self.scores_max = list(argrelmin(self.func_scores)[0])

        self.scores_maxs_pair = subsequences([0]+self.scores_max+[len(self.func_scores)],2)

        self.phrases = []
        self.phrases_span = []
        for l,r in self.scores_maxs_pair:

            self.phrases.append({k:type for k,type in zip(parad_chain.labels[l:r],[True if s==min(self.func_scores[l:r]) else False for s in self.func_scores[l:r]])})

            self.phrases_span.append({l:s for l,s in zip(parad_chain.labels[l:r],parad_chain.spans[l:r])})

        self.phrases_nodes = {"".join(phr.keys()):(list((phr.values()))[0][0],list((phr.values()))[-1][-1]) for phr in self.phrases_span}

    def __getitem__(self, index:str):
        return self.types[index]
    
    def plot_scores(self, show_mean = False, show_probs = False):

        if show_probs == True:

            scaling_factor = np.max(self.probs)/np.max(self.func_scores)

        fig = plot_scatter_line(
            x=self.indexes,
            y=self.func_scores,
            trace_name= "Func. Scores",
            title=" ".join(self.labels),
            xaxis_title='Paradigm',
            yaxis_title='Functional Score',
            add_trace = (self.indexes, np.array(self.probs)/scaling_factor, "Probs.") if show_probs else None
            )

        if show_mean == True:
            score_mean = np.mean(self.func_scores)
            fig.add_shape(
                type="line",
                x0=0, x1 = self.len-1, y0 = score_mean, y1 = score_mean, 
                line=dict(
                    color="LightSeaGreen",
                    width=2,
                    dash="dash",
                ))
        


        return fig



from . import normalizers
from . import pre_tokenizers
from . import processors
from . import post_processors

# from ...util import if_none_disable

# TODO: unable to import this from ...util
def if_none_disable(x):
    if x == None:
        x = "disable"
    return x

class SLG_Tokenizer:
    """
    """

    def __init__(self,config) -> None:

        # self.config = {k:v if v!= None else "disable" for k,v in config.__dict__.items()}
        self.config = config
        

        # TODO: There should be a better way to convert config "None" into "disable"
        
        if isinstance(self.config.SLG_normalizer,list):
            self.normalizer = normalizers.Sequence(self.config.SLG_normalizer)
        else:
            self.normalizer = eval(f"normalizers.{if_none_disable(self.config.SLG_normalizer)}()")
        
        self.pre_tokenizer = eval(f"pre_tokenizers.{if_none_disable(self.config.SLG_pre_tokenizer)}()")
        self.is_pretokenized = False if self.pre_tokenizer == pre_tokenizers.PreTokenizer else True
        self.processor = eval(f"processors.{if_none_disable(self.config.SLG_processor)}()")
        self.post_processor = eval(f"post_processors.{if_none_disable(self.config.SLG_post_processor)}()")

    def encoder(self,input_string):
        pass

    def decoder(self,input_string):
        pass

    def __call__(self,chain):
        chain.norm = self.normalizer.normalize(chain.input)
        chain.pre_tokens = self.pre_tokenizer.pre_tokenize(chain.norm)
        chain.processor = self.processor.process(chain.pre_tokens, chain.semiotic, is_pretokenized = self.is_pretokenized)
        chain.tree_tokens = self.post_processor.post_process(chain.processor)
        chain.tokens = sorted([token for token in chain.tree_tokens if token.position != None], key= lambda x: x.position)

        chain.len = len(chain.tokens)
        chain.labels = [token.label for token in chain.tokens]
        chain.probs = [token.prob for token in chain.tokens]
        pass


# EXTERNAL MODULES



# Functions for the Chain class



#%%

# ###############################
# # 00 - GENERAL                #
# ###############################

# def execute_procedure(step_list:list):
#     for step in step_list:
#         print(f"\nExecuting step:\t{step}")
#         print(f"===============\t{'='*len(step)}\n")
#         importlib.import_module("."+step,package="analyses")
#     print("\nProcedure Finished!\n")


# ###############################
# # 01 - SEGMENTATION FUNCTIONS #
# ###############################

# # 01.01 Building Vocabulary #
# #############################

# def load_corpus(
#     filename:str,
#     length = None,
#     ):
#     print("Loading Corpus")
#     start = time.perf_counter()
#     with open(f"{paths.corpus}{filename}.txt", "r") as f:
#         chain = f.read(length)
#     finish = time.perf_counter()
#     print(f"Corpus loaded in {round(finish - start,2)} secs\n")
#     return chain

# def find_best_pair(chain_spaced):
#     pairs = Counter()
#     pre_units = chain_spaced.split()
#     for i in range(len(pre_units) - 1):
#         pairs[pre_units[i], pre_units[i + 1]] += 1
#     return pairs.most_common()[0][0]

# def agglutinate_chain(pair, chain_spaced):
#     bigram = re.escape(" ".join(pair))
#     p = re.compile(r"(?<!\S)" + bigram + r"(?!\S)")
#     new_chain = p.sub("".join(pair), chain_spaced)
#     return new_chain

# def count_units_in_chain(chain_spaced):
#     vocab = Counter()
#     chain_units = chain_spaced.split()
#     for unit in chain_units:
#         vocab[unit] += 1
#     return dict(vocab.most_common())

# def build_vocab(
#     chain:str,
#     voc_length: int,
#     inter_save = [],
#     save_finalQ = True,
#     filename = "voc_default",
#     resumeQ = False,
#     corpus_name = "corp_deefault",
#     courpus_length = None,
# ):
#     if resumeQ:
#         initial_chain = chain.replace(" ", "")
#     else:
#         initial_chain = chain
#     spaced_chain = " ".join(chain)
#     alpha = count_units_in_chain(spaced_chain)
#     alpha_len = len(alpha)
#     print(f"Length of initial alphabet: {alpha_len}")
#     if resumeQ:
#         print("Resuming chain...")
#         spaced_chains = chain
#         print("Chain resumed")
#     if voc_length > 0:
#         print("Enter loop")
#         for i in range(voc_length):
#             start = time.perf_counter()
#             best_pair = find_best_pair(spaced_chain)
#             spaced_chain = agglutinate_chain(best_pair, spaced_chain)
#             finish = time.perf_counter()
#             print(
#                 f"{i}: {''.join(best_pair)} - Computed in {round(finish - start,2)} secs"
#             )
#             if i in inter_save:
#                 vocab = count_units_in_chain(spaced_chain)
#                 print(
#                     f"Saving intermediate result. Vocabulary length: {len(vocab)-alpha_len}"
#                 )
#                 output_file = (
#                     f"voc_A_{corpus_name}_{len(vocab)-alpha_len}_inter"
#                 )
#                 util.dict2csv(vocab, output_file, paths.vocabularies)

#                 output_resume = f"{corpus_name}_resume"
#                 util.str2txt(spaced_chain, output_resume, paths.scratch)
#                 util.str2txt(
#                     f"Input corpus: {corpus_name}\nCorpus Lenght: {courpus_length}\nParallel: {parallelQ}\nNumbers of Cores: {n_cores}\nVocabulary length (so far): {len(vocab)-alpha_len}",
#                     f"{corpus_name}_resume_info",
#                     paths.scratch,
#                 )
#     print("Collecting frequencies of terms")
#     vocab = count_units_in_chain(spaced_chain)
#     if save_finalQ:
#         util.dict2csv(vocab, filename, paths.vocabularies)
#     return vocab

# def find_best_pair_par(chain_spaced):
#     pairs = Counter()
#     pre_units = chain_spaced.split()
#     for i in range(len(pre_units) - 1):
#         pairs[pre_units[i], pre_units[i + 1]] += 1
#     return dict(
#         pairs.most_common()[:100]
#     )  # Looking only on top 100 pre_units of each par list, for efficiency

# def build_vocab_par(
#     chain: str,
#     voc_length: int,
#     n_cores=4,
#     inter_save=[],
#     resumeQ = False,
#     save_finalQ=False,
#     filename = "voc_p_default",
#     corpus_name = "corp_deefault",
#     courpus_length = None,
#     ):
#     if resumeQ:
#         initial_chain = chain.replace(' ','').replace('\n','')
#     else:
#         initial_chain = chain
#     print("Spacing chain...")
#     spaced_chains = [" ".join(chain_part) for chain_part in util.partition(initial_chain, n_cores)]
#     print(f"Chain spaced")
#     alphas = util.multiprocessing(count_units_in_chain, spaced_chains)
#     alpha = dict(functools.reduce(operator.add, [Counter(voc) for voc in alphas]).most_common())
#     apha_len = len(alpha)
#     print(f'Length of initial alphabet: {apha_len}')
#     if resumeQ:
#         print("Resuming chain...")
#         spaced_chains = chain.split('\n')
#         print('Chain resumed')
#     if voc_length > 0:
#         print("Enter loop")
#         for i in range(voc_length):
#             start = time.perf_counter()
#             top_pairs_par = util.multiprocessing(find_best_pair_par, spaced_chains)
#             best_pair = functools.reduce(
#                 operator.add, [Counter(top_pairs) for top_pairs in top_pairs_par]
#             ).most_common()[0][0]
#             spaced_chains = util.multiprocessing(
#                 functools.partial(agglutinate_chain, best_pair), spaced_chains
#             )
#             finish = time.perf_counter()
#             print(
#                 f"{i}: {''.join(best_pair)} - Computed in {round(finish - start,2)} secs"
#             )
#             if i in inter_save:
#                 vocabs = util.multiprocessing(count_units_in_chain, spaced_chains)
#                 vocab = dict(
#                     functools.reduce(
#                         operator.add, [Counter(voc) for voc in vocabs]
#                     ).most_common()
#                 )
#                 print(f"Saving intermediate result. Vocabulary length: {len(vocab)-apha_len}")
#                 output_file = f"voc_A_{corpus_name}_p_{len(vocab)-apha_len}_inter"
#                 util.dict2csv(vocab, output_file, paths.vocabularies)

#                 output_resume = f"{corpus_name}_p_resume"
#                 util.str2txt('\n'.join(spaced_chains), output_resume, paths.scratch)
#                 util.str2txt(f'Input corpus: {corpus_name}\nCorpus Lenght: {courpus_length}\nParallel: {parallelQ}\nNumbers of Cores: {n_cores}\nVocabulary length (so far): {len(vocab)-apha_len}',f"{corpus_name}_p_resume_info",paths.scratch)
#     print('Collecting frequencies of terms')
#     vocabs = util.multiprocessing(count_units_in_chain, spaced_chains)
#     vocab = dict(
#         functools.reduce(operator.add, [Counter(voc) for voc in vocabs]).most_common()
#     )
#     if save_finalQ:
#         util.dict2csv(vocab, filename, paths.vocabularies)
#     return vocab

# def build_vocabulary(
#     chain,
#     printQ = True,
#     parallelQ = True,
#     n_cores = 4,
#     voc_len = 10,
#     inter_results = False,
#     resumeQ = False,
#     save_finalQ = False,
#     filename = "voc_default",
#     corpus_name = "corp_deefault",
#     courpus_length = None,
# ):
#     start = time.perf_counter()
#     print(f"Building Vocabulary")
#     if printQ:
#         print("Print is on")
#     else:
#         print("Print is off")
#     if parallelQ:
#         print(f"Method: Parallel - N° of Cores: {n_cores}")
#         voc = build_vocab_par(chain, voc_len, n_cores, inter_save=inter_results, save_finalQ=save_finalQ, resumeQ = resumeQ, filename = filename, corpus_name = corpus_name, courpus_length = courpus_length)
#     else:
#         print(f"Method: Sequential")
#         voc = build_vocab(chain, voc_len, inter_save=inter_results, save_finalQ=save_finalQ,resumeQ = resumeQ, filename = filename, corpus_name = corpus_name, courpus_length = courpus_length)
#     finish = time.perf_counter()
#     print(f"Vocabulary built in {round(finish - start,2)} secs")
#     return voc


# # 01.02 Segment Sentences #
# ###########################

# def load_vocabulary(
#     filename,
#     ):
#     # print(f"Loading Vocabulary ({os.path.basename(filename)})")
#     with open(f"{paths.vocabularies}{filename}.csv", "r") as f:
#         csv_reader = csv.reader(f)
#         voc_A = Counter()
#         for line in csv_reader:
#             voc_A[line[0]] = int(line[1])
#         voc_A = dict(voc_A)
#     # print('Done!\n')
#     return voc_A

# def load_test_sents(
#     filename
#     ):
#     print(f"Loading Test Sentences ({filename})")
#     with open(f"{paths.sentences}{filename}.txt", "r") as f:
#         sents = []
#         for line in f.readlines():
#             sents.append(line.rstrip())
#     print('Done!\n')
#     return sents

# # test_sent = "i have made my plans and i must stick to them"


# # def random_sent(sents):
# #     return sents[random.randrange(len(sents))]





# def tree_encoder(tree_data):
#     seg_tree = nx.DiGraph()
#     seg_tree.add_edges_from(tree_data)
#     string = list(seg_tree.nodes())[0][0]
#     gdf = {n: i for i, n in enumerate(seg_tree.nodes())}
#     gs = [(gdf[h], gdf[t]) for h, t in seg_tree.edges()]
#     node_interval = [n[1] for n in seg_tree.nodes()]
#     encoded_tree = (string, node_interval, gs)
#     return encoded_tree

# def forest_encoder(trees):
#     print("Encoding forest...")
#     start = time.perf_counter()
#     encoded_forest = util.multithreading(tree_encoder, trees)
#     finish = time.perf_counter()
#     print(f"Forest of {len(trees)} trees encoded in {round(finish-start,2)} secs")
#     return encoded_forest

# def save_forest(forest, filename):
#     print("Saving segmented trees")
#     encoded_forest = forest_encoder(forest)
#     util.list2csv(encoded_forest, filename, paths.segmentations)
#     print(f"Segmented chains saved as: {filename}")
#     return ''


# def segment_chainlist(
#     sents:list,
#     voc:dict,
#     seg_type = "sq",
#     weight_f = rankOnly,
#     zipf_factor = .135,
#     sample_size = 100,
#     randomQ = False,
#     parallelQ = True,
#     saveQ = False,
#     filename = "seg_default",
#     ):
#         assert seg_type in {"sq", "tr"}, "The segmentation type (seg_type) should be either 'sq' or 'tr'"

#         print("Computing segmented sequences")
#         start = time.perf_counter()
#         if sample_size==None:
#             sample_size = len(sents)
#         if randomQ:
#             chain_list = random.sample(sents,min(sample_size,len(sents)))
#         else:
#             chain_list = sents[:sample_size]
        
#         voc_rank = vocabulary_rank(voc,zipf_factor)

#         if seg_type == "sq":
#             print("Sequences always computed sequentially (no parallelism)")
#             segs = [" ".join(chain2seq(chain, voc, weight_f,voc_rank, zipf_factor)) for chain in chain_list]
            
#             print(f"Segmented chains (sequences) computed in:{time.perf_counter()-start} secs.")

#             if saveQ:
#                 print("Saving segmented sequences")
#                 util.str2txt("\n".join(segs),f"{seg_type}_{filename}",paths.segmentations)
#                 print(f"Segmented chains saved as: {filename}")
#             else:
#                 print("Segmented chains not saved.")
                
#         elif seg_type == "tr":
#             if parallelQ:
#                 print("Computing trees in parallel")
#                 segs = util.multiprocessing(functools.partial(chain2tree, voc=voc, weight_f=weight_f, v_rank = voc_rank, zipf_factor = zipf_factor), chain_list)
#             else:
#                 print("Computing trees sequentially")
#                 segs = [chain2tree(chain, voc, weight_f,voc_rank, zipf_factor) for chain in chain_list]

#             print(f"Segmented chains (trees) computed in:{time.perf_counter()-start} secs.")

#             if saveQ:
#                 save_forest(segs,f"{seg_type}_{filename}")
#             else:
#                 print("Segmented trees not saved.")


#         print("Done!\n")

#         return segs

# def plot_tree(tree_data, red=None, grey=None, fname=f"{paths.segment_graphs}seg_graph"):
#     seg_tree_graph = gv.Digraph(name=fname)
#     seg_tree = nx.DiGraph()
#     seg_tree.add_edges_from(tree_data)
#     tree_nodes_list = [(str(i), l) for l, i in list(seg_tree.nodes)]
#     for node in tree_nodes_list:
#         seg_tree_graph.node(
#             *node, color="white", fontsize="30", fontname="garamond"
#         )  # style="filled", color="grey")
#     seg_tree_graph.attr("edge", color="slategrey")
#     seg_tree_graph.edges([(str(p[1]), str(c[1])) for p, c in list(seg_tree.edges)])
#     if grey == None:
#         pass
#     else:
#         for node in [(str(i), l) for l, i in grey]:
#             if node in tree_nodes_list:
#                 seg_tree_graph.node(
#                     *node, style="dashed", color="grey", fontsize="30", fontname="garamond"
#                 )

#     if red == None:
#         pass
#     else:
#         for node in [(str(i), l) for l, i in red]:
#             if node in tree_nodes_list:
#                 seg_tree_graph.node(*node, color="red", fontsize="30", fontname="garamond")
#     return seg_tree_graph


# # 01.03 Extract Orthogonals #
# #############################


# def load_seqs(filename, n_start = None, n_end = None):
#     with open(f"{paths.segmentations}{filename}.txt", "r") as f:
#         seqs = []
#         for line in f.readlines():
#             seqs.append(line.rstrip())
#     return seqs[n_start:n_end]

# def tree_decoder(encoded_tree):
#     string = encoded_tree[0]
#     nodes_dict = {i: (string[n[0] : n[1]], n) for i, n in enumerate(encoded_tree[1])}
#     tree_data = [(nodes_dict[h], nodes_dict[t]) for h, t in encoded_tree[2]]

#     return tree_data

# def forest_decoder(encoded_trees):
#     print("Decoding trees...")
#     start = time.perf_counter()
#     decoded_trees = util.multiprocessing(tree_decoder, encoded_trees)
#     finish = time.perf_counter()
#     print(f"{len(decoded_trees)} trees decoded in {round(finish-start,2)} secs")
#     return decoded_trees

# def load_trees(filename, n_start = None, n_end = None, decodeQ = True):

#     loaded_trees = util.csv2list(filename, paths.segmentations, n_start, n_end)

#     if decodeQ:
#         output_forest = forest_decoder(loaded_trees)
#     else:
#         output_forest = loaded_trees
    
#     return output_forest

# def load_segs(
#     filename,
#     n_start = None,
#     n_end = None
#     ):
#     seg_type = filename[:2]
#     assert seg_type in {"sq","tr"}, "The type of the segmentation file does not seem to match the possible data types ('sq' and 'tr')"

#     start = time.perf_counter()
#     print(f"Loading segmentations from: {filename}")
#     if seg_type=="sq":
#         seg_type_name = "sequences"
#         segs = load_seqs(filename, n_start, n_end)
#     elif seg_type=="tr":
#         seg_type_name = "trees"
#         segs = load_trees(filename, n_start, n_end)

#     finish = time.perf_counter()
#     print(f"{len(segs)} {seg_type_name} loaded in {round(finish-start,2)} secs")
#     print('Done!\n')
#     return segs

# def extract_bigrams_from_chains(list_of_chains):
#     bigrams = []
#     for chain in list_of_chains:
#         bigrams += util.subsequences(chain.split(), 2)
#     bg_dict = dict(Counter(bigrams).most_common())
#     return bg_dict


# def extract_siblings(tree_data, Llen = [], Rlen = []):

#     tree = nx.DiGraph()
#     tree.add_edges_from(tree_data)

#     neighbors = [tuple([n[0] for n in tree.neighbors(node)]) for node in tree.nodes()]
#     siblings = []
#     for p in neighbors:
#         if  p != ():
#             siblings.append(p)
#     return siblings

# def bulk_extract_siblings(forest):
    
#     siblings = util.multithreading(extract_siblings,forest)
#     siblings_flat = list(itertools.chain(*siblings))

#     siblings_dict = Counter(siblings_flat)

#     return dict(siblings_dict.most_common())


# def save_orthogonals(orthogonals, filename, freq_min = 0):
#     if not os.path.isdir(paths.orthogonals):
#         os.makedirs(paths.orthogonals)
#     with open(f"{paths.orthogonals}{filename}_{freq_min}.csv", "w") as nf:
#         for key, value in orthogonals.items():
#             if value >= freq_min:
#                 nf.write(f"{key[0]},{key[1]},{value}\n")


# def extract_orthogonals(
#     list_of_segs,
#     saveQ = False,
#     filename = "ortho_default",
#     freq_min = 0,
#     ):
#     print('Extracting orthogonals...')
#     seg_type = type(list_of_segs[0])
#     assert seg_type in {str,list}, "Format of segmentations not recognized"

#     start = time.perf_counter()
#     if seg_type == str:
#         print('Extracting bigrams from sequences...')
#         orthos = extract_bigrams_from_chains(list_of_segs)
#     elif seg_type == list:
#         print('Extracting siblings from trees...')
#         orthos = bulk_extract_siblings(list_of_segs)
#     finish = time.perf_counter()
#     print(f"Orthogonals of {len(list_of_segs)} {'sequences' if seg_type==str else 'trees'} extracted in {round(finish-start,2)} secs")
#     if saveQ:
#         if type(freq_min) == list:
#             for n in freq_min:
#                 save_orthogonals(orthos,filename,n)
#                 print(f"Orthogonals saved as {filename}_{n}")
#         else:
#             save_orthogonals(orthos,filename,freq_min)
#             print(f"Orthogonals saved as {filename}_{freq_min}")
#     else:
#         print("Orthogonals not saved")
#     print('Done\n')
#     return orthos


# # #########################
# # # 02 - TYPING FUNCTIONS #
# # #########################


# def load_orthogonality_data(
#     filename,
#     freq_min = 0,
#     ):
#     print(f"Loading data... ({os.path.basename(filename)})")
#     print(f"Loading orthogonals of frequency equal or higher than: {freq_min}")
#     with open(f"{paths.orthogonals}{filename}.csv", "r") as f:
#         csv_reader = csv.reader(f)
#         biGrams_all = Counter()
#         for line in csv_reader:
#             if int(line[-1]) >= freq_min:
#                 biGrams_all[tuple([line[i] for i in range(len(line)-1)])] = int(line[-1])
#     print(f"Orthogonals final length: {len(biGrams_all)}")
#     print(f"Done!\n")
#     return dict(biGrams_all.most_common())

# def build_terms_contexts(
#     orthogonals,
#     hand_picked_terms=[],
#     hand_picked_contexts=[],
#     trim_terms=None,
#     trim_contexts=None,
#     symmetricQ=False):
#     # TODO: There should be a much faster way of constructing this matrix by extracting the info of the dic and directly building a csr_matrix out of it
#     print("Building Terms and Contexts...")
#     terms_dict = util.marginalize(orthogonals,"left")
#     contexts_dict = util.marginalize(orthogonals,"right")
#     if symmetricQ in ('i','u'):
#         terms_context_union = Counter(terms_dict)+Counter(contexts_dict)
#         terms = contexts = [k for k,v in terms_context_union.most_common()]
#         if symmetricQ == "i":
#             terms_contexts_i = set(list(terms_dict.keys())).intersection(set(list(contexts_dict.keys())))
#             terms = contexts = [t for t in terms if t in terms_contexts_i]
#         if trim_terms==None:
#             print('No Trim')
#         else:
#             print(
#                 f"Trim terms and contexts: {trim_terms}"
#             )
#             terms = contexts = terms[:trim_terms]

#     else:
#         terms = [k for k,v in Counter(terms_dict).most_common()]
#         contexts = [k for k,v in Counter(contexts_dict).most_common()]
#         if trim_terms==None:
#             print('No Term Trim')
#         else:
#             print(
#                 f"Trim terms: {trim_terms}"
#             )
#             terms = terms[:trim_terms]
#         if trim_contexts==None:
#             print('No Context Trim')
#         else:
#             print(
#                 f"Trim Contexts: {trim_contexts}"
#             )
#             contexts = contexts[:trim_contexts]

#     if len(hand_picked_terms) > 0:
#         terms = hand_picked_terms
#     if len(hand_picked_contexts) > 0:
#         contexts = hand_picked_contexts
#     print(f"Terms length: {len(terms)}")
#     print(f"Contexts length: {len(contexts)}")
#     print("Done\n")
#     return [terms, contexts]


# def mm_no_modif(contexts, orthogonals, term):
#     return [
#         orthogonals.get((term, context),0) for context in contexts
#     ]

# def matrix_maker(
#     terms,
#     contexts,
#     orthogonals,
#     measure = mm_no_modif):
#     results = util.multithreading(functools.partial(measure, contexts, orthogonals), terms)
#     return results

# def build_term_context_matrix(
#     terms,
#     contexts,
#     orthogonals,
#     normalizeQ = False):
#     print("Building oR Matrix...")
#     start = time.perf_counter()
#     if normalizeQ:
#         orthogonals = util.normalize_dict(orthogonals)
#     matrix = csr_matrix(matrix_maker(terms,contexts,orthogonals))
#     finish = time.perf_counter()
#     print(f"Term-Context Matrix built in {round(finish-start,2)} secs.\n")
#     return matrix

# def build_pmi_matrix(
#     term_context_matrix,
#     type = "pmi",
#     alpha = .75,
#     normalizeQ = False,
#     ):

#     print("Computing PMI Matrix...")
#     print(f"Type: {type}")
#     if "s" in type:
#         print(f"Smoothing (alpha): {alpha}")
#     start = time.perf_counter()
#     pmi_matrix = util.pmi(term_context_matrix,alpha=alpha,type_pmi=type)
#     finish = time.perf_counter()
#     if normalizeQ:
#         print("Normalizing Matrix")
#         pmi_matrix = (1/(pmi_matrix.sum()))*pmi_matrix
#     print(f"PMI Matrix built in {round(finish-start,2)} secs.")
#     print("Done\n")
#     return pmi_matrix

# def cut_thres(row, thres):
#     ortho_index = (row > thres) * 1
#     return ortho_index

# def cut_mean(row, thres):
#     if len(row.data) == 0:
#         ortho_index = row
#     else:
#         mean_cut = row.data.mean()
#         if mean_cut == row.data.max():
#             mean_cut = mean_cut / 2
#         mean_cut = row.data.mean()*thres
#         ortho_index = (row > mean_cut) * 1
#     return ortho_index

# def build_bin_matrix(
#     matrix,
#     cut_func = "cut_thres",
#     thres = 1e-5,
#     parallelQ = False
#     ):
#     """
#     In general, parallelizing in this case is slower than the sequential method (and multiprocessing slower than multithreading). This should still tried out. Maybe other ways of parallelizing are possible.
#     """
#     print(f"Binarizing Matrix...")
#     print(f"Cut Function: {cut_func}")
#     cut_func_eval = eval(cut_func)
#     start = time.perf_counter()
#     if parallelQ:
#         print("Method: Parallel")
#         disc_rows = util.multithreading(
#             functools.partial(cut_func_eval, thres=thres), matrix
#         )
#     else:
#         print("Method: Sequential")
#         disc_rows = []
#         for row in matrix:
#             disc_row = cut_func_eval(row, thres)
#             disc_rows.append(disc_row)
#     finish = time.perf_counter()
#     print(f"Binary matrix built in {round(finish-start,2)} secs.\n")
#     return vstack(disc_rows).astype(np.bool)

# def reduce_matrix(
#     matrix,
#     reduce_thres = 2):
#     """
#     Remove rows from binary matrix that have less units than a given threeshold (min_len)
#     """
#     print(f"Reducing matrix. Threshold: {reduce_thres}")
#     ones_factor = np.ones(matrix.shape[1],dtype=np.int8)
#     len_test = matrix.dot(ones_factor)
#     indices = np.where(len_test>=reduce_thres)
#     result = matrix[indices]
#     print(f"New reduced shape: {result.shape}")
#     return result

# def indiv_intersection(set_of_sets, reduce_thres, indiv_set):
#     inter_set = {indiv_set.intersection(s) for s in set_of_sets}
#     inter_set = {s for s in inter_set if len(s)>=reduce_thres}
#     inter_set = inter_set-set_of_sets
#     return inter_set

# def powerset_M(
#     matrix,
#     reduce_thres = 2,
#     bound = 20,
#     parallel=False,
#     ):
#     """
#     Takes a Boolean matrix and computes a matrix whose rows are the conjunction of all possible combinations of the rows of the original matrix (the conjunction of all the elements of the powerset of the set of original rows).
    
#     Rows with strictly less components than the integer defined as "reduce_thres" are disregarded.
    
#     The option "bound" establishes a limit of cardinality for the elements of the powerset to be computed.

#     Parallelism seems to slow down the computation

#     The final order of the rows in the matrix is not handled
#     """
#     start = time.perf_counter()



#     def row_sets_to_matrix(row_as_set:tuple):
#         r_index = row_as_set[0]
#         r_set = row_as_set[1]
#         matrix_row = [[r_index,c] for c in r_set]
#         return matrix_row

#     def sets2matrix(set_of_sets, width, parallel=False):
#         """
#         Converts a set of frozen sets into a sparse matrix in which each frozenset is a row and its elements are indices of columns. The order of the rows is not handled.
#         """
#         if parallel:
#             print("Building matrix in parallel")
#             csr_indices = util.multithreading(row_sets_to_matrix,enumerate(list(set_of_sets)))
#         else:
#             csr_indices = [row_sets_to_matrix(row) for row in enumerate(list(set_of_sets))]
#         csr_indices = np.array(util.flatten(csr_indices))
#         data = np.ones(len(csr_indices))
#         ps_M = csr_matrix((data, (csr_indices[:,0], csr_indices[:,1])), shape=(len(set_of_sets),width), dtype=bool)
#         return ps_M

#     print(f"Input size:\t{matrix.shape}")
#     matrix = reduce_matrix(matrix, reduce_thres)

#     print("Computing the conjunction of all the subsets of rows from matrix")
#     print(f"Parallel:\t{parallel}")

#     n = 1
#     # M_length = matrix.shape[0]
    
#     print("Transforming matrix into set of sets...")
#     matrix_dict = defaultdict(list)
#     for i, j in zip(*matrix.nonzero()):
#         matrix_dict[i].append(j)

#     matrix_as_sets = {frozenset(i) for i in matrix_dict.values()}


#     n_subset_M = matrix_as_sets
#     collect_n_subsets = matrix_as_sets

#     while n<bound+1:

#         n += 1
#         print(f"Subsets of size: {n}")

#         start_st = time.perf_counter()
        
#         if parallel:
#             n_subset_M = util.multithreading(functools.partial(indiv_intersection, n_subset_M, reduce_thres), matrix_as_sets)
#         else:
#             n_subset_M = [indiv_intersection(n_subset_M, reduce_thres,row) for row in matrix_as_sets]     

#         print(f"****\t Step computed in {round(time.perf_counter()-start_st,2)} secs. \t****")

#         start_st = time.perf_counter()
#         # n_subset_M = util.sp_unique(csr_matrix(vstack(n_subset_M_n)))
#         n_subset_M = set.union(*n_subset_M)
#         print(f"****\t Step computed in {round(time.perf_counter()-start_st,2)} secs. \t****")


#         start_st = time.perf_counter()
#         n_subset_M = n_subset_M-collect_n_subsets
#         print(f"****\t Step computed in {round(time.perf_counter()-start_st,2)} secs. \t****")
#         print(f"New:\t{len(n_subset_M)}")


#         if len(n_subset_M)==0:
#             print(f"\nEmpty subsets from subsets greater than {n-1} elements.\n")
#             break
#         else:
#             start_st = time.perf_counter()
#             collect_n_subsets = collect_n_subsets.union(n_subset_M)
#             print(f"Stacked:\t{len(collect_n_subsets)}")
#             print(f"****\t Step computed in {round(time.perf_counter()-start_st,2)} secs. \t****")

#         print("\n")

#     print("Transforming set of sets into matrix...")
#     start_st = time.perf_counter()
#     ps_M = sets2matrix(collect_n_subsets, matrix.shape[1],parallel=parallel)
#     print(f"Matrix built in {round(time.perf_counter()-start_st,2)} secs.")


#     finish = time.perf_counter()
#     print(f"Powerset matrix computed in {round(finish-start,2)} secs.")
#     print(f"Final size:\t{ps_M.shape}")
#     print("=================================\n")

#     return ps_M

# def powerset_M_parallel(
#     matrix_list:list,
#     reduce_thres = 2,
#     bound = 20,
#     ):
#     """
#     Computes the powerset of several matrices in parallel. Matrices need to be given in a list.
#     """
#     print("Computing Power Set Matrices...\n")
#     st = time.perf_counter()
#     matrix_list_ps = util.parallel_processes(
#         [
#             [functools.partial(powerset_M,
#             reduce_thres=reduce_thres,
#             bound=bound,
#             parallel=False),matrix]
#             for matrix in
#             matrix_list]
#     )
#     fin = time.perf_counter()
#     print(f"Powerset of {len(matrix_list)} matrices computed in parallel in {round(fin-st,2)} secs.\n")
#     return matrix_list_ps


# def orthogonal_row(row, ooQ):
#     if row.sum() == 0:
#         biorthos_vec = csr_matrix((1, ooQ.shape[1]), dtype=np.int8)
#     else:
#         indices = row.indices
#         ortho_rows = ooQ[indices, :]
#         biorthos_vec = ortho_rows.min(axis=0)
#     return biorthos_vec

# def biorthogonal(
#     oQ,
#     ooQ,
#     parallelQ=False
#     ):
#     # Parallelization is slower here
#     """
#     Given two corresponding binary matrixes of orthogonality (ex: to the Right [oQ] and back to the Left of that Right [ooQ]), this function computes the biorthogonal terms for each row of oQ (NB: each row might correspond to the orthogonals of multiple terms if initial matrix results form powerset_M). The resulting matrix is no longer a term-context matrix, but a terms-term (the class of terms out of which the bi-orthogonal types are composed belong to the same set of terms as the class of terms generating it)
#     """
#     print("Computing Biorthogonal Matrix...")
#     start = time.perf_counter()
#     if parallelQ:
#         print("Method: Parallel")
#         row_list = [r for r in oQ]
#         biorthos = util.multithreading(
#             functools.partial(orthogonal_row, ooQ=ooQ), row_list
#         )
#     else:
#         print("Method: Sequential")
#         biorthos = []
#         for row in oQ:
#             biorthos_vec = orthogonal_row(row, ooQ)
#             biorthos.append(biorthos_vec)
#     finish = time.perf_counter()
#     print(f"Biorthogonal Matrix computed in {round(finish-start,2)} secs.\n")
#     return csr_matrix(vstack(biorthos),dtype=np.bool)

# def biorthogonal_parallel(
#     list_pairs_matrices:list
# ):
#     """
#     Computes the biorthogonal of several (pairs of) matrices in parallel. Pairs of matrices need to be given in a list.
#     """
#     print("Computing biorthogonal matrices...")
#     bo_M = util.parallel_processes(
#         [
#             [biorthogonal,o,oo]
#             for o,oo in
#             list_pairs_matrices
#             ]
#     )
#     return bo_M

# def bo_Types(
#     terms,
#     contexts,
#     bo_M,
#     o_M,
#     min_len=2,
#     saveQ = False,
#     filename = "bo_o_",
#     ):
#     """
#     bo_M and o_M should be of shape (n,len(terms)) and (n,len(contexts))
#     if filter_sT = True, it filters out those couples whose ortho type is included in that of another couple of the list for the same bo type
#     """
#     print('Building bo Types List...')
#     assert bo_M.shape[0]==o_M.shape[0], "ERROR: bo_M and o_M are not of same length"
#     start = time.perf_counter()
#     types_list = []
#     for index in range(bo_M.shape[0]):
#         types = (
#             tuple(sorted([terms[i] for i in bo_M[index].indices])),
#             tuple(sorted([contexts[i] for i in o_M[index].indices]))
#         )
#         types_list.append(types)

#     # First possible structural decission: Exclude orthogonal relations involving types of length smaller than a given n (min_len) (typically 2)
#     # Possible objection: There could be categories with unmarked members (for instance the singular in English, with "s" as mark of the plural, or "th" as mark of ordinal)
#     print(f"Filtering to types of length equal or greater than {min_len}")
#     types_list = [(l,r) for l,r in types_list if min(len(l), len(r)) >= min_len]

#     finish = time.perf_counter()
#     print(f"Types List built in {round(finish-start,2)} secs.\n")

#     if saveQ:
#         util.list2csv(
#             types_list,
#             filename,
#             paths.types
#             )
#     print(f"Types table saved as: {filename}")

#     return types_list


# def build_types_parallel(
#     quad_t_c_bo_o_list:list,
#     min_len = 2,
#     saveQ = False,
#     filename = "bo_o_",
#     ):
#     """
#     Computes the biorthogonal of several matrices in parallel. lists of [terms, contexts, bo-matrix and o-matrix] need to be given in a list.
#     """
#     bo_oR_ = util.parallel_processes(
#         [
#             [functools.partial(bo_Types,min_len=min_len, saveQ=saveQ, filename=filename+"_"+str(n+1)),t,c,bo,o]
#             for n,(t,c,bo,o) in
#             enumerate(quad_t_c_bo_o_list)
#         ]
#     )
#     return bo_oR_

# def load_type_list(
#     filename,
#     n_start = None,
#     n_end = None,
#     ):
#     def csv2T_list(filename: str, directory, n_start=None, n_end=None):
#         with open(directory+filename, "r") as f:
#             csv_reader = csv.reader(f)
#             my_list = [
#                 (ast.literal_eval(lT), ast.literal_eval(rT))
#                 for lT,rT in util.take(n_end, csv_reader)
#             ]
#         return my_list[n_start:n_end]
#     files = [fn for fn in os.listdir(paths.types) if filename in fn]
#     type_lists = [
#         csv2T_list(fn,paths.types,n_start,n_end)
#         for fn in files
#     ]
#     return type_lists



# def subtype_rels(type_list, raw = True):
#     """
#     Builds a list of tuples (i,j) of types, where j is a subtype of i
#     Set raw = True for raw list of types (list of tuples). If False, the function takes the first element of each item of the list (used for building graphs)
#     """
#     # type_list = set(type_list)
#     lattice_edges = []
#     for i in type_list:
#         for j in type_list:
#             if raw:
#                 if set(i).issubset(set(j)) and i != j:
#                     lattice_edges.append((j, i))
#             else:
#                 if set(i[0]).issubset(set(j[0])) and i[0] != j[0]:
#                     lattice_edges.append((j, i))
#     return lattice_edges

# def filter_biortho_subtypes(types_list):
#     """
#     Given a list of couples of bo-o types, it filters out those couples whose ortho type is included in that of another couple of the list for the same bo type
#     """
#     types_list_filter_dict = defaultdict(list)
#     for boT,oT in types_list:
#         types_list_filter_dict[boT].append(oT)
#     reduced_types_list = []
#     for boT in (i for i in types_list_filter_dict.keys() if len(i)>1):
#         oT_indiv = util.delete_duplicates(types_list_filter_dict[boT])
#         oT_st = subtype_rels(oT_indiv)
#         oT_bot = [b for t,b in oT_st]
#         reduced_types_list_indiv = [(boT,oT) for oT in oT_indiv if oT not in oT_bot]
#         reduced_types_list.extend(reduced_types_list_indiv)
#     return reduced_types_list

# def build_ortho_graph(
#     bo_types_list,
#     side = "R",
#     ortho=True,
#     subtype_filter=True,
#     filter_sT = False,
#     verbose=True
#     ):
    
#     print('Building Graph...') if verbose else None

#     if filter_sT:
#         print("Filtering orthogonal subtypes")
#         bo_types_list = filter_biortho_subtypes(bo_types_list)

#     print('Building orthogonal relations') if verbose else None

#     if side == "R":
#         boT_side = "L"
#         oT_side = "R"
#     else:
#         boT_side = "R"
#         oT_side = "L"

#     ortho_rels = [
#         ((l, "bo", boT_side), (r, "o", oT_side))
#         for l, r in bo_types_list        
#     ]

#     # Second structural decission: Exclude orthogonal relations whose terms don't involve subtyping relations with other terms in their respective sides
#     # Possible objection: The search for types through orthogonality is not directly related to subtyping (ex: what should be the subtyping structure to which "ed,ing,s" belongs?)

#     if subtype_filter:
#         print('Building L, R, o and bo subtyping filters') if verbose else None

#         struct_elements_boT = []
#         for i,j in subtype_rels([l for l, r in ortho_rels], raw = False):
#             struct_elements_boT += [i,j]
#         struct_elements_boT = set(struct_elements_boT)

#         struct_elements_oT = []
#         for i,j in subtype_rels([r for l, r in ortho_rels], raw = False):
#             struct_elements_oT += [i,j]
#         struct_elements_oT = set(struct_elements_oT)

#         print('Filtering orthogonalities to subtypes structures') if verbose else None

#         struct_ortho_rels = [
#             (i, j)
#             for i, j in ortho_rels
#             if i in struct_elements_boT and j in struct_elements_oT
#         ]
#     else:
#         struct_ortho_rels = ortho_rels

#     print('Constructing transitive reduction of lattices') if verbose else None

#     tred_graph_boT = nx.DiGraph()
#     tred_graph_boT.add_edges_from(subtype_rels([l for l, r in struct_ortho_rels], raw = False))
#     tred_graph_boT = nx.algorithms.dag.transitive_reduction(tred_graph_boT)

#     tred_graph_oT = nx.DiGraph()
#     tred_graph_oT.add_edges_from(subtype_rels([r for l, r in struct_ortho_rels], raw = False))
#     tred_graph_oT = nx.algorithms.dag.transitive_reduction(tred_graph_oT)

#     print('Merging boT and oT graphs') if verbose else None
#     complete_graph = functools.reduce(
#         nx.compose, [tred_graph_boT, tred_graph_oT]
#     )

#     print('Adding orthogonal edges') if verbose else None
#     if side == "L":
#         struct_ortho_rels = [(r,l) for l,r in struct_ortho_rels]
#     if ortho:
#         complete_graph.add_edges_from(struct_ortho_rels, edge_type="ortho")
#     print("Done!\n") if verbose else None
#     return complete_graph

# def compose_graphs(list_of_graphs: list):
#     return functools.reduce(
#         nx.compose, list_of_graphs
#     )

# def make_label(label: tuple, LRN="N"):
#     """
#     Given a list of tuples corresponding to Types, it outputs strings for names of those
#     Types in the graph, inserting linebrakes if the string is too long.
#     Options for making L or R types with 'LRN = ': 'L', 'R', 'LR'.
#     """
#     if len(label) < 6:
#         lab = str(label)
#     elif len(label) > 15:
#         lab_cut = str(label[:15])
#         groups = str(lab_cut).split(" ")
#         lab = "\n".join(
#             [
#                 " ".join(groups[:5]),
#                 " ".join(groups[6:10]),
#                 " ".join(groups[11:15]) + "... " + "[+" + str(len(label) - 15) + "]",
#             ]
#         )
#     else:
#         n = int(len(label) / 2)
#         groups = str(label).split(" ")
#         lab = "\n".join([" ".join(groups[:n]), " ".join(groups[n:])])
#     if LRN == "L":
#         return lab.replace("'", "").replace("(", "").replace(")", "") + " /"
#     elif LRN == "R":
#         return "\\\\ " + lab.replace("'", "").replace("(", "").replace(")", "")
#     elif LRN == "LR":
#         return "\\\\ " + lab.replace("'", "").replace("(", "").replace(")", "") + " /"
#     else:
#         return lab.replace("'", "").replace("(", "").replace(")", "")

# def simple_lattice_graph(
#     lattice_nx_graph,
#     fname="LatGraph",
#     fdir=paths.lattice_graphs,
#     colors_list=[
#         "burlywood2",
#         "blanchedalmond",
#         "grey80",
#         "slategrey",
#     ],
#     font="RobotoMono-Thin"
# ):
#     print("Plotting graph...")

#     lattice_g = gv.Digraph(name=fname, directory=fdir, strict=True)

#     lattice_g.attr("node", fontname = font)
#     lattice_g.attr("edge", color="grey80")

#     lattice_g.edges(
#         [(str(i), str(j)) for i, j in lattice_nx_graph.edges()]
#     )

#     for node in lattice_nx_graph.nodes:
#         lattice_g.node(
#             str(node),
#             make_label(node),
#             style="filled",
#             colorscheme = "OrRd9",
#             color=len(node),
#             shape="oval",
#             #orientation="0" if node[-1] == "L" else "90",
#             #margin="0,0" if node[0] in LR_intersect else "0.2,0.2",
#         )
#     print("Finished!")
#     return lattice_g

# def plot_lattice_graph(raw_parads,filter_single_nodes = False):
#     lattice_rels = subtype_rels(raw_parads)
#     tred_graph = nx.DiGraph()
#     tred_graph.add_edges_from(lattice_rels)
#     if not filter_single_nodes:
#         tred_graph.add_nodes_from(raw_parads)
#     tred_graph = nx.algorithms.dag.transitive_reduction(tred_graph)

#     return simple_lattice_graph(tred_graph)

# def lattice_graph(
#     complete_nx_graph,
#     fname="Graph",
#     fdir=paths.lattice_graphs,
#     colors_list=[
#         "burlywood2",
#         "blanchedalmond",
#         "grey80",
#         "slategrey",
#     ],
#     dependenciesQ=True,
#     LRQ = True,
#     font="RobotoMono-Thin"
# ):
#     print("Plotting graph...")
#     LR_intersect = set([n[0] for n in complete_nx_graph.nodes() if n[-1] == 'L']) & set([n[0] for n in complete_nx_graph.nodes() if n[-1] == 'R'])

#     color_switch = {
#         ("o","L") : 1,
#         ("bo","R") : 0,
#         ("bo","L") : 0,
#         ("o","R") : 1,
#         ("obo","L") : 2,
#         ("obo","R") : 2,
#     }

#     Complete_LR = gv.Digraph(name=fname, directory=fdir, strict=True)

#     # top_bottom = {("⊤",), ("⊥",)}
#     Complete_LR.attr("node", fontname = font)
#     Complete_LR.attr("edge", color="grey80")

#     Complete_LR.edges(
#         [(str(i), str(j)) for i, j, d in complete_nx_graph.edges(data=True) if d == {}]
#     )

#     for node in complete_nx_graph.nodes:
#         # if node[1] == "o":
#         Complete_LR.node(
#             str(node),
#             make_label(node[0], LRN=node[2]),
#             style="filled",
#             color=colors_list[color_switch[node[1:3]]],
#             shape="doubleoctagon" if node[0] in LR_intersect else "cds",
#             orientation="0" if node[-1] == "L" else "90",
#             margin="0,0" if node[0] in LR_intersect else "0.2,0.2",
#         )

#     if dependenciesQ:
#         Complete_LR.attr("edge", arrowhead="none", color="darksalmon")
#         Complete_LR.edges(
#             [
#                 (str(i), str(j))
#                 for i, j, d in complete_nx_graph.edges(data=True)
#                 if d != {}
#             ]
#         )
#     print("Finished!")
#     return Complete_LR

# def bigram_types(bigram, types_lists, separate_lists=False):
#     possible_types = []
#     for list in types_lists:
#         ortho_types = [(lT, rT) for lT, rT in list if bigram[0] in lT and bigram[1] in rT]
#         if separate_lists:
#             possible_types.append(ortho_types)
#         else:
#             possible_types.extend(ortho_types)
#     return possible_types


# # Draft functions
# from transformers import pipeline

# unmasker = pipeline('fill-mask', model='distilbert-base-uncased',top_k=10)

# def sent_paradigm(sentence,thres=0):
#     sent_list = sentence.split()
#     sent_mask = [" ".join([token if n!=i else "[MASK]" for n,token in enumerate(sent_list)]) for i in range(len(sent_list))]
#     parads = [{i['token_str']:i['score'] for i in unmasker(sent) if i['score']>thres and i['token_str'] not in {
#         ".",":",",","…","'","’","′",'"',"•",";","`","-","“","...","?","!","/","&","–"
#         } } for sent in sent_mask]

#     return (parads)

# def extract_parads(sents_sample,thres=.01): 

#     parads = Counter()
#     i=0
#     print_i = [i*10 for i in range(len(sents_sample))]
#     for sent in sents_sample:
#         i+=1
#         if i in print_i:
#             print(f"{i}/{len(sents_sample)}")
#         sent_parad, probability_mass = sent_paradigm(sent,thres=thres,prob=False)
#         for p in sent_parad:
#             if len(p)>1:
#                 parads[tuple(sorted(p))]+=1
#     return parads


# def simple_lattice_graph(
#     lattice_nx_graph,
#     fname="LatGraph",
#     fdir=paths.lattice_graphs,
#     colors_list=[
#         "burlywood2",
#         "blanchedalmond",
#         "grey80",
#         "slategrey",
#     ],
#     font="RobotoMono-Thin"
# ):

#     lattice_g = gv.Digraph(name=fname, directory=fdir, strict=True)

#     lattice_g.attr("node", fontname = font)
#     lattice_g.attr("edge", color="grey80")

#     lattice_g.edges(
#         [(str(i), str(j)) for i, j in lattice_nx_graph.edges()]
#     )
#     max_node = max([node[1] for node in lattice_nx_graph.nodes])
#     for node in lattice_nx_graph.nodes:
#         node_1_9_ranking = max(1,int(node[1]*9/max_node))
#         lattice_g.node(
#             str(node),
#             make_label(node[0]),
#             style="filled",
#             colorscheme = "oranges9",
#             color = str(node_1_9_ranking),
#             fontcolor = "white" if node_1_9_ranking>7 else "black",
#             shape= "doubleoctagon" if node[1]==max_node else "oval",
#             #orientation="0" if node[-1] == "L" else "90",
#             #margin="0,0" if node[0] in LR_intersect else "0.2,0.2",
#         )
#     return lattice_g

# def plot_lattice_graph(raw_parads,filter_single_nodes = False):
#     lattice_rels = subtype_rels(raw_parads,raw=False)
#     tred_graph = nx.DiGraph()
#     tred_graph.add_edges_from(lattice_rels)
#     if not filter_single_nodes:
#         tred_graph.add_nodes_from(raw_parads)
#     tred_graph = nx.algorithms.dag.transitive_reduction(tred_graph)

#     return simple_lattice_graph(tred_graph)


# For the moment only AdamWeightDecay is implemented. If other alternatives are considered, then they should be imported here from transformers
from transformers import BertConfig, TFBertForMaskedLM, BertTokenizer, AdamWeightDecay, DataCollatorForLanguageModeling #, pipeline
import datasets
from .paradigm import Paradigmatizer, ParadigmChain

import tensorflow as tf
import numpy as np
from os import path, makedirs

from ..util import dict2json

# adapted from https://stackoverflow.com/questions/68038820/how-to-save-the-best-model-of-each-epoch-with-transformers-bert-in-tensorflow/68042600#68042600
class model_per_epoch(tf.keras.callbacks.Callback):
    def __init__(self, model, filepath):
        self.filepath = filepath
        self.model = model
    def on_epoch_end(self, epoch, logs=None):
        v_loss = logs.get('val_loss')
        name= str(epoch) +'-' + str(v_loss)[:str(v_loss).rfind('.')+3] + '.h5'
        file_id=path.join(self.filepath, name)
        self.model.save_weights(file_id, overwrite=True)
        print(f"\nSLG: Checkpoint saved as: {name}")

class Paradigmatic:
    def __init__(self,semiotic) -> None:
        self.config = semiotic.config.paradigmatic
        self.split_rate = semiotic.config.corpus.split_rate
        self.model_config = semiotic.config.paradigmatic.model_config
        self.max_len = semiotic.config.syntagmatic.model_max_length
        self.tensor_imp = semiotic.config.general.tensor_implementation
        self.cpu_count = semiotic.config.system.cpu_count
        self.multiple_gpus = semiotic.config.system.multiple_gpus
        self.path = semiotic.paths.paradigms
        self.syntagmas_path = semiotic.paths.syntagmas
        self.model_config_path = self.path / "config.json"

        if self.config.huggingface_pretrained == None:
            self.hf_tokenizer = None
            self.hf_model = None
        else:
            self.hf_tokenizer = BertTokenizer.from_pretrained(self.config.huggingface_pretrained)
            self.hf_model = TFBertForMaskedLM.from_pretrained(self.config.huggingface_pretrained)


        self.dataset = semiotic.corpus.dataset

        self.tokenizer = semiotic.syntagmatic.bert_tokenizer
        self.build_tokenized_dataset = semiotic.syntagmatic.build

        self.bert_config = BertConfig(
            vocab_size = self.tokenizer.vocab_size,
            hidden_size = self.model_config["hidden_size"],
            num_hidden_layers = self.model_config["num_hidden_layers"],
            num_attention_heads = self. model_config["num_attention_heads"],
            intermediate_size = self. model_config["intermediate_size"],
            hidden_act = self. model_config["hidden_act"],
            hidden_dropout_prob = self. model_config["hidden_dropout_prob"],
            attention_probs_dropout_prob = self. model_config["attention_probs_dropout_prob"],
            max_position_embeddings = self. model_config["max_position_embeddings"],
            type_vocab_size = self. model_config["type_vocab_size"],
            initializer_range = self. model_config["initializer_range"],
            layer_norm_eps = self. model_config["layer_norm_eps"],
            pad_token_id = self. model_config["pad_token_id"],
            position_embedding_type = self. model_config["position_embedding_type"],
            use_cache = self. model_config["use_cache"],
            classifier_dropout = self. model_config["classifier_dropout"],
        )

        # If model exists and config.load_pretrain==True, load pretrained model, otherwise load an empty model
        if self.tensor_imp == "tf":
            if self.config.load_pretrained and path.exists(self.path / "tf_model.h5") and path.exists(self.model_config_path):
                if self.multiple_gpus:
                    mirrored_strategy = tf.distribute.MirroredStrategy()
                    with mirrored_strategy.scope():
                        self.model = TFBertForMaskedLM.from_pretrained(self.path / "tf_model.h5",
                            config = self.model_config_path
                            )
                else:
                    self.model = TFBertForMaskedLM.from_pretrained(
                        self.path / "tf_model.h5",
                        config = self.model_config_path
                        )
                print(f"SLG [I]: Paradigmatizer loaded from disk")
            else:
                self.model = TFBertForMaskedLM(self.bert_config)
                print(f"SLG [I]: Empty paradigmatizer model loaded")

        elif self.tensor_imp == "pt":
            raise Exception(f"SLG: Models other than TensorFlow are not yet implemented in SemioLog. tensor_implementation is set to {self.tensor_imp}. Please set it to 'tf'")
        else:
            raise Exception(f"SLG: Models other than TensorFlow are not yet implemented in SemioLog. tensor_implementation is set to {self.tensor_imp}. Please set it to 'tf'")

        self.optimizer = eval(
            f"{self.config.optimizer['optimizer']}        (learning_rate={self.config.optimizer['learning_rate']}, weight_decay_rate={self.config.optimizer['weight_decay']})"
        )
        
        self.data_collator = DataCollatorForLanguageModeling(
            tokenizer = self.tokenizer,
            mlm = True,
            mlm_probability = self.config.mask_probability,
            pad_to_multiple_of = None,
            return_tensors = self.tensor_imp
        )

        # self.unmasker = pipeline('fill-mask', model = self.config.model,top_k = self.config.top_k)
        self.paradigmatizer = Paradigmatizer(self.model,self.bert_tokenizer,semiotic.syntagmatic.tokenizer.decode)

    # This looks inelegant. There should be a way to do this in a more intelligent way
    def bert_tokenizer(
        self,
        input,
        add_special_tokens = True,
        padding = True,
        truncation = False,
        # max_length = self.max_len,
        stride = 0,
        is_split_into_words = False,
        pad_to_multiple_of = None,
        # return_tensors = self.tensor_imp,
        return_token_type_ids = True,
        return_attention_mask = True,
        return_overflowing_tokens = False,
        return_special_tokens_mask = False,
        return_offsets_mapping = False,
        return_length = False,
        verbose = True,
        ):
        output = self.tokenizer(
            text = input,
            add_special_tokens = add_special_tokens,
            padding = padding,
            truncation = truncation,
            max_length = self.max_len,
            stride = stride,
            is_split_into_words = is_split_into_words,
            pad_to_multiple_of = pad_to_multiple_of,
            return_tensors = self.tensor_imp,
            return_token_type_ids = return_token_type_ids,
            return_attention_mask = return_attention_mask,
            return_overflowing_tokens = return_overflowing_tokens,
            return_special_tokens_mask = return_special_tokens_mask,
            return_offsets_mapping = return_offsets_mapping,
            return_length = return_length,
            verbose = verbose,
        )
        return output
    
    def build(
        self,
        dataset = None,
        epochs = None,
        load_tokenized = None,
        n_sents = None,
        checkpoints = False,
        save = None,
        min_token_length = 2,
        max_token_length = 256,
        # checkpoint_weights = None,
        ):
        
        if dataset == None:
            dataset = self.dataset
        
        if epochs == None:
            epochs = self.config.training_epochs

        if load_tokenized == None:
            load_tokenized = self.config.load_tokenized

        if n_sents == None:
            n_sents = self.config.n_sents

        if save == None:
            save = self.config.save

        # Following Huggingface, no loss and metrics are provided

        print("SLG [I]: Compiling model")
        self.model.compile(
            optimizer = self.optimizer,
            # optimizer=tf.keras.optimizers.Adam(learning_rate=5e-5),
            # loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
            # metrics=tf.metrics.SparseCategoricalAccuracy(),
            )

            
        # Commented out because of bugs on loading weights. Checkpoints are now saved as models, not as weights. They should be loaded differently
        # if checkpoint_weights!=None: 
        #     if path.isfile(self.path / f"checkpoints/{checkpoint_weights}"):
        #         self.model.built = True
        #         self.model.load_weights(self.path / f"checkpoints/{checkpoint_weights}", by_name = True, skip_mismatch = True)
        #         print(f"SLG [I]: Checkpoint weights loaded from {checkpoint_weights}.")
        #     else:
        #         print(f"SLG [W]: Checkpoint file {checkpoint_weights} not found. Weights will not be loaded.")

        if load_tokenized and path.exists(self.syntagmas_path / "tokenized"):
            
            tokenized_datasets = datasets.load_from_disk(self.syntagmas_path / "tokenized")
            print("SLG [I]: Tokenized dataset loaded from disk")

            if n_sents !=None:
                
                #TODO: Maybe if both sizes are bigger than the respective datasets, it would be best not to perform select at all

                if n_sents > tokenized_datasets["train"].num_rows:
                    print(f"SLG [paradigmatic - W]: n_sents greater than num_rows in tokenized_datasets.train. Keeping tokenized_datasets.train full size")
                    n_sents = tokenized_datasets["train"].num_rows
                if int(n_sents*(1/self.split_rate[0])*self.split_rate[1]) > tokenized_datasets["dev"].num_rows:
                    print(f"SLG [paradigmatic - W]: n_sents greater than num_rows in tokenized_datasets.train. Keeping tokenized_datasets.train full size")
                    n_sents_dev = tokenized_datasets["dev"].num_rows
                else:
                    n_sents_dev = int(n_sents*(1/self.split_rate[0])*self.split_rate[1])


                tokenized_datasets = datasets.DatasetDict({
                    "train":tokenized_datasets["train"].select(range(n_sents)),
                    "dev": tokenized_datasets["dev"].select(range(n_sents_dev))
                    })


        else:

            tokenized_datasets = self.build_tokenized_dataset(
                None,
                save_tokenized = False,
                n_sents = n_sents,
                return_tokenized = True,
            )

        print(f"SLG [I]: Filtering rows of length < {min_token_length} and > {max_token_length}")
        tokenized_datasets = datasets.DatasetDict({
            "train":tokenized_datasets["train"].filter(lambda example: min_token_length<=len(example['input_ids'])<=max_token_length),
            "dev": tokenized_datasets["dev"].filter(lambda example: min_token_length<=len(example['input_ids'])<=max_token_length)
            })

        print("SLG [I]: Building train set")
        train_set = tokenized_datasets["train"].to_tf_dataset(
            columns = ["attention_mask", "input_ids", "labels"],
            shuffle = self.config.input_sets["shuffle"],
            batch_size = self.config.input_sets["batch_size"],
            collate_fn = self.data_collator,
        )

        print("SLG [I]: Building validation set")
        validation_set = tokenized_datasets["dev"].to_tf_dataset(
            columns = ["attention_mask", "input_ids", "labels"],
            shuffle = self.config.input_sets["shuffle"],
            batch_size = self.config.input_sets["batch_size"],
            collate_fn = self.data_collator,
        )

        if checkpoints:
            checkpoint_filepath = self.path / "checkpoints"
            if not path.isdir(checkpoint_filepath):
                makedirs(checkpoint_filepath)
            model_checkpoint_callback=[model_per_epoch(self.model, checkpoint_filepath)]  
        else:
            model_checkpoint_callback = None

        print("SLG [I]: Start training...\n")  
        self.history = self.model.fit(
            train_set,
            validation_data = validation_set,
            epochs = epochs,
            callbacks = model_checkpoint_callback
        )
        print("SLG [I]: Training finished")

        if self.config.save:

            self.model.save_pretrained(save_directory = self.path)
            print("\nSLG [I]: Model saved.")

            dict2json(self.history.history, "history", self.path)
            print("SLG [I]: Training history saved.")

        print("SLG [I]: Model built!")


from collections import Counter, defaultdict
import csv

import socket
socket_name = socket.gethostname()
if any(name in socket_name for name in {"Gianni","vpn","Berenice"}):
    from tqdm.notebook import tqdm, trange
else:
    from tqdm.auto import tqdm, trange
    
import regex as re
from os import makedirs, listdir
from os.path import isfile, isdir
from functools import reduce
import operator
from joblib import Parallel, delayed
import time

from tokenizers import normalizers
from datasets import Dataset,logging
logging.set_verbosity_error()
# from .syntagmatic import NormalizeSLG

from . import util
from .syntagmatic import Syntagmatic, NormalizeSLG, Normalize_w_spacesSLG # needed
from .paradigmatic import Paradigmatic

# TODO: Solve version as global variable
slg_version = "0.2.3"
    
class Vocabulary:
    
    def __init__(self,semiotic):
        
        #TODO: Is there another way than loading the corpus (or the semiotic) here?
        self.corpus = semiotic.corpus
        self.semiotic = semiotic
        
        self.name = semiotic.name
        self.path = semiotic.paths.vocabulary
        self.config = semiotic.config.vocabulary
        self.model = self.config.model
        self.cpu_count = semiotic.config.system.cpu_count
        self.thres_ = 0.000001 #TODO: This should go to the configs
        
        self.merges = None
        self.encode = None
        self.freq = None
        self.alpha = None
        
        self.decode = None
        
        self.len = None
        self.freq_mass = None
        self.prob = None

        

        # Load HF normalizer
        
        # #TODO: The elif condition on the string SLG is sort of a hack (needed due to non standard declaration of custom normalizer). There should be a more elegant way

        config_normalizer = semiotic.config.syntagmatic.normalizer
        if config_normalizer != None:
            if isinstance(config_normalizer,list):
                normalizer = normalizers.Sequence(
                    [eval(f"normalizers.{norm}()") for norm in config_normalizer]
                    )
            elif "SLG" in config_normalizer:
                normalizer = eval(f"{config_normalizer}")
            else:
                normalizer = eval(f"normalizers.{config_normalizer}()")
        
            self.normalizer = normalizer.normalize_str
        
        else:
            # self.normalizer = lambda x: x
            self.normalizer = None


    @property
    def thres(self):
        return self.thres_

    def from_file(self,path = None):
        if path == None:
            path = self.path


        filenames = [(path / fn) for fn in ["merges.txt","vocab.json","freq.json","alpha.json"]]
        
        for filename in filenames:
            if not isfile(filename):
                return print(f"SLG [W]: {filename} does not exist.\nVocabulary will not be loaded from disk.\n")
        
        self.merges = [tuple(merge.split()) for merge in util.load_file(path / "merges.txt")[1:]] # The first line needs to be stripped
        self.encode = util.json2dict("vocab",path)
        self.freq = util.json2dict("freq",path)
        self.alpha = util.json2dict("alpha",path)

        self.decode = {i:k for k,i in self.encode.items()}
        
        self.len = len(self.encode)
        self.freq_mass = sum(self.freq.values())
        self.char_mass = sum(self.alpha.values())
        self.prob = {k:v/self.freq_mass for k,v in self.freq.items()}
        print(f"SLG [I]: Vocabulary loaded from disk")
    
        # TODO: implement better automatic loading of all files in ngram

        if isdir(self.path / "ngrams"):
            ngram_files = sorted([f for f in listdir(self.path / "ngrams") if isfile(self.path / f"ngrams/{f}") and f[-4:]=="json"], key=lambda x: int(x.split(".")[0].split("_")[0]))

            self.ng1 = nGram(from_dict=self.alpha)
            if ngram_files!=[]:
                for f in tqdm(ngram_files):
                    print(f"Loading nGram file {f}...", end="\r")
                    setattr(self, f"ng{f.split('.')[0]}", nGram(from_file = self.path / f"ngrams/{f}"))
                print(f"SLG [I]: nGrams loaded from disk ({ngram_files})")


    def __repr__(self) -> str:
        return f"Voc({self.freq})"

    def __str__(self) -> str:
        return str(self.freq)

    def __getitem__(self, item):
         return self.prob[item]

    def head(self,size=10):
        return list(self.freq.items())[:size]
    
    def tail(self,size=10):
        return list(self.freq.items())[-size:]

    def alphabetic(self):
        pass

    def keys(self):
        pass

    def values(self):
        pass

    def build(
        self,
        corpus = None,
        model = None,
        vocab_size = None,
        special_tokens = None,
        save = False,
        save_step = None,
        progress_bar = True,
        resume_merges = False,
        parallel = False,
        corpus_length = None,
        keep_in_memory = False, #HF argument in dataset shard and map
        ):

        if corpus == None:
            corpus = self.name
        
        if model == None:
            model = self.model
        
        if vocab_size == None:
            vocab_size = self.config.size

        if special_tokens == None:
            special_tokens = self.config.special_tokens

        if corpus_length == None:
            corpus_length = self.corpus.train_len

        if save == True and save_step != None:
            saveQ = True
            
            if not isdir(self.path):
                makedirs(self.path)
        else:
            saveQ = False

        if model == "SLG":
            def pre_process(corpus_chunk:Dataset, normalizer):

                # Normalize
                
                if normalizer == None:
                    chain_zip = corpus_chunk
                else:
                    # chain_zip = normalizer(corpus_chunk)
                    chain_zip = corpus_chunk.map(lambda sent: {"text": normalizer(sent["text"])}, keep_in_memory=keep_in_memory)

                chain_zip = "".join(chain_zip["text"])
                
                # Build list of pairs
                chain_zip = list(zip(chain_zip,chain_zip[1:]))
                # Create a lookup table of all the positions where a pair appears in a corpus
                pair_pos = defaultdict(set)
                for i,k in list(enumerate(chain_zip)):
                    pair_pos[k].add(i)
                # From the previous lookup table, create another lookup table of the frequency of each pair (given by the size of the set of its positions)
                pair_len = Counter()
                for k,pos in pair_pos.items():
                    pair_len[k] = len(pos)
                
                return (chain_zip, pair_pos, pair_len)

            def process_best_pair(chain_zip, pair_pos, best_pair):
                chain_zip_len = len(chain_zip)
                pair_len_delta = Counter()

                for i in pair_pos[best_pair]:
                    # Skip iteration if position corresponds to a modified set of positions during the iteration. This can happen if there is overlap of pairs, such as "000", where ("0","0") has itself as right pair. # Note that, due to unordered implementation of sets, this entails a lack of systematicity in overlapping cases: "000" can be counted randomly as ("00","0") or ("0","00").
                    if chain_zip[i]!=best_pair:
                        continue
                    ## merge best pair with left unit
                    left_pair_i = i-1
                    while left_pair_i>=0 and chain_zip[left_pair_i] == None: # if left pair is within chain limits but empty (= None) because already merged previously, shift to the left
                        left_pair_i -= 1
                    if left_pair_i>-1: # proceed only if a left pair was found on the left
                        # Remove from left pair positions, the current position (of the pair to be merged)
                        left_pair = chain_zip[left_pair_i]
                        # Skip update of left_pair position set if left_pair = best_pair, to avoid modification of iterating set. This can happen if there is overlap of pairs. No consequences on final result (right?) since right after the loop, the key corresponding to the best pair is deleted, and chain_zip is indeed updated so the problematic cases can be captured at the beginning of the loop.
                        if left_pair != best_pair:
                            pair_pos[left_pair].discard(left_pair_i)
                        new_pair = (left_pair[0],"".join(best_pair)) # construct new left pair
                        
                        # update the list of pairs
                        chain_zip[left_pair_i] = new_pair
                        # add new pair (if non existing) and its position to the pair_pos lookup table
                        pair_pos[new_pair].add(left_pair_i)
                        # update the counts in the pair_len lookuptable
                        pair_len_delta[left_pair] -= 1
                        pair_len_delta[new_pair] += 1


                    ## merge best pair with right unit.
                    # Code is symmetric to left_pair but on the right. Comments are omitted
                    right_pair_i = i+1
                    while right_pair_i<chain_zip_len and chain_zip[right_pair_i] == None:
                        right_pair_i += 1
                    if right_pair_i<chain_zip_len:
                        right_pair = chain_zip[right_pair_i]
                        if right_pair != best_pair:
                            pair_pos[right_pair].discard(right_pair_i)
                        new_pair = ("".join(best_pair), right_pair[1])

                        chain_zip[right_pair_i] = new_pair
                        pair_pos[new_pair].add(right_pair_i)
                        pair_len_delta[right_pair] -= 1
                        pair_len_delta[new_pair] += 1


                    # Empty best pair position in list of pairs
                    chain_zip[i] = None

                # Remove best pair from lookuptables
                del pair_pos[best_pair]

                return (chain_zip, pair_pos, pair_len_delta)

            def compute_freq(chain_zip):
                # Collect the first component of the pairs
                freq = [pair[0] for pair in chain_zip if pair != None]
                i = -1
                # Add the last component of the last pair
                while chain_zip[i]==None:
                    i -= 1
                freq.append(chain_zip[i][-1])
                # Count the units of the resulting (decoupled) chain list
                freq = Counter(freq)
                return freq
            

            if parallel:
                # TODO: The chunks limits could be improved (in particular, if corpus_length is very small compared to cpu_count, last chunks may be empty. It shouldn't be a problem for large corpus_length)
                
                print(f"SLG: Computing in parallel. CPU count: {self.cpu_count}")
                
                corpus_chunks = [self.corpus.train.shard(self.cpu_count,n, contiguous=True, keep_in_memory = keep_in_memory) for n in range(self.cpu_count)]

                with Parallel(n_jobs=self.cpu_count, require='sharedmem') as parallel_pool:
                    print("SLG [I]: Starting parallel jobs.")
                    print("SLG [I]: Normalize and jobs data...")
                    start = time.time()
                    jobs_data = parallel_pool(delayed(pre_process)(chunk,self.normalizer) for chunk in corpus_chunks)

                    pair_len_global = reduce(operator.add,[pair_len for chain_zip, pair_pos, pair_len in jobs_data])

                    # When pair_len_global has more than 1 max, the first encountered is chosen, introducing possible discrepancies between implementations (because each choice modifies global statistics). However, multiple max is less likely to appear in big corpora and relatively small vocabularies, and mostly at the tail of vocabularies (ie. low frequencies), so the impact of this divergence is expected to be marginal.
                    best_pair, best_pair_len = max(pair_len_global.items(), key=operator.itemgetter(1))
                    
                    merges = [" ".join(best_pair)]
                    print(f"... computed in {time.time()-start} secs.\n")

                    print("SLG [I]: Build alphabet...")
                    start = time.time()
                    alphabet = Counter()
                    for (l,r),v in pair_len_global.items():
                        alphabet[l] += v
                    # In extreme cases, right characters of pairs might not be left characters. If there are such chars, they're added with freq 1
                    left_out_chars = {r for l,r in pair_len_global.keys()}-alphabet.keys()
                    if len(left_out_chars)>0:
                        print(f"SLG: Adding characters: {left_out_chars}")
                        for char in left_out_chars:
                            alphabet[char] += 1
                    print(f"... computed in {time.time()-start} secs.\n")

                    alpha_len = len(alphabet)
                    special_tokens_len = 0 if special_tokens == None else len(special_tokens)
                    
                    print(f"SLG: Alphabet Size: {alpha_len}")
                    print(f"SLG: Special Tokens Size: {special_tokens_len}")
                    
                    if vocab_size<0:
                        voc_final_length = alpha_len + special_tokens_len + abs(vocab_size)
                    else:
                        voc_final_length = vocab_size

                    delta_voc = voc_final_length - alpha_len - special_tokens_len

                    print(f"SLG: Terms to compute: {delta_voc}\n")

                    print("SLG: Enter loop")

                    t = trange(delta_voc, disable = not progress_bar)
                    for _ in t:
                        t.set_description(f"Pair: {best_pair}, {best_pair_len}")
                        t.refresh()

                        jobs_data = parallel_pool(delayed(process_best_pair)(chain_zip, pair_pos, best_pair) for chain_zip, pair_pos, pair_len_delta in jobs_data)

                        for chain_zip, pair_pos, pair_len_delta in jobs_data:
                            pair_len_global.update(pair_len_delta)

                        # Remove best_pair from pair_len
                        del pair_len_global[best_pair]

                        # When pair_len_global has more than 1 max, the first encountered is chosen, introducing possible discrepancies between implementations (because each choice modifies global statistics). However, multiple max is less likely to appear in big corpora and relatively small vocabularies, and mostly at the tail of vocabularies (ie. low frequencies), so the impact of this divergence is expected to be marginal.
                        best_pair, best_pair_len = max(pair_len_global.items(), key=operator.itemgetter(1))

                        merges.append(" ".join(best_pair))

                        if saveQ == True:
                            voc_partial_len = alpha_len + special_tokens_len + _ + 1
                            if voc_partial_len % save_step == 0 and voc_partial_len != voc_final_length:

                                print("SLG [I]: Saving intermediate results...")
                                start = time.time()
                                freqs = parallel_pool(delayed(compute_freq)(chain_zip) for chain_zip, pair_pos, pair_len_delta in jobs_data)
                                freq = reduce(operator.add, freqs)

                                vocabulary = freq.most_common()
                                
                                if special_tokens != None:
                                    vocabulary = vocabulary + [(token,0) for token in special_tokens]
                                
                                self.merges = merges
                                self.encode = {k:i for i,(k,v) in enumerate(vocabulary)}
                                self.freq = dict(vocabulary)
                                self.alpha = dict(alphabet.most_common())
                                step_path = self.path /  "checkpoints/" + str(voc_partial_len)
                                self.save(step_path)
                                print(f"... computed in {time.time()-start} secs.")
                                print(f"SLG [I]: Intermediate vocabulary saved to {step_path}\n")

                    print("SLG [I]: Compute freq...")
                    start = time.time()
                    freqs = parallel_pool(delayed(compute_freq)(chain_zip) for chain_zip, pair_pos, pair_len_delta in jobs_data)
                    freq = reduce(operator.add, freqs)
                    print(f"... computed in {time.time()-start} secs.\n")
            
            else:
                #TODO: Sequential computing not completely tested
                print("SLG [I]: Computing sequentially")
                print("SLG [I]: Normalize and jobs data...")
                start = time.time()
                corpus_chain = self.corpus.train[:corpus_length]
                chain_zip, pair_pos, pair_len_global = pre_process(corpus_chain,self.normalizer)

                # When pair_len_global has more than 1 max, the first encountered is chosen, introducing possible discrepancies between implementations (because each choice modifies global statistics). However, multiple max is less likely to appear in big corpora and relatively small vocabularies, and mostly at the tail of vocabularies (ie. low frequencies), so the impact of this divergence is expected to be marginal.
                best_pair, best_pair_len = max(pair_len_global.items(), key=operator.itemgetter(1))
                
                merges = [" ".join(best_pair)]
                print(f"... computed in {time.time()-start} secs.\n")

                print("SLG [I]: Build alphabet...")
                start = time.time()
                alphabet = Counter()
                for (l,r),v in pair_len_global.items():
                    alphabet[l] =+ v
                # In extreme cases, right characters of pairs might not be left characters. If there are such chars, they're added with freq 1
                left_out_chars = {r for l,r in pair_len_global.keys()}-alphabet.keys()
                if len(left_out_chars)>0:
                    print(f"SLG: Adding characters: {left_out_chars}")
                    for char in left_out_chars:
                        alphabet[char] =+ 1
                print(f"... computed in {time.time()-start} secs.\n")

                alpha_len = len(alphabet)
                special_tokens_len = 0 if special_tokens == None else len(special_tokens)
                
                print(f"SLG: Alphabet Size: {alpha_len}")
                print(f"SLG: Special Tokens Size: {special_tokens_len}")
                
                if vocab_size<0:
                    voc_final_length = alpha_len + special_tokens_len + abs(vocab_size)
                else:
                    voc_final_length = vocab_size

                delta_voc = voc_final_length - alpha_len - special_tokens_len
                
                print(f"SLG: Terms to compute: {delta_voc}\n")

                print("SLG: Enter loop")

                t = trange(delta_voc, disable = not progress_bar)
                for _ in t:
                    t.set_description(f"Pair: {best_pair}, {best_pair_len}")
                    t.refresh()

                    chain_zip, pair_pos, pair_len_delta = process_best_pair(chain_zip, pair_pos, best_pair)

                    # Remove best_pair from pair_len
                    pair_len_global.update(pair_len_delta)

                    del pair_len_global[best_pair]

                    # When pair_len_global has more than 1 max, the first encountered is chosen, introducing possible discrepancies between implementations (because each choice modifies global statistics). However, multiple max is less likely to appear in big corpora and relatively small vocabularies, and mostly at the tail of vocabularies (ie. low frequencies), so the impact of this divergence is expected to be marginal.
                    best_pair, best_pair_len = max(pair_len_global.items(), key=operator.itemgetter(1))

                    merges.append(" ".join(best_pair))
                    # print(f"... computed in {time.time()-start} secs.\n")

                    if saveQ == True:
                        voc_partial_len = alpha_len + special_tokens_len + _ + 1
                        if voc_partial_len % save_step == 0 and voc_partial_len != voc_final_length:

                            print("SLG: Saving intermediate results...")
                            start = time.time()
                            freq = compute_freq(chain_zip)

                            vocabulary = freq.most_common()
                            
                            if special_tokens != None:
                                vocabulary = vocabulary + [(token,0) for token in special_tokens]
                            
                            self.merges = merges
                            self.encode = {k:i for i,(k,v) in enumerate(vocabulary)}
                            self.freq = dict(vocabulary)
                            self.alpha = dict(alphabet.most_common())
                            step_path = self.path / str(voc_partial_len)
                            self.save(step_path)
                            print(f"... computed in {time.time()-start} secs.")
                            print(f"SLG: Intermediate vocabulary saved to {step_path}\n")
                
                print("SLG [I]: Compute freq...")
                start = time.time()
                freq = compute_freq(chain_zip)
                print(f"... computed in {time.time()-start} secs.\n")

        elif model == "SLG_WL":

            freq = Counter(" ".join(self.corpus.train["text"]).split())
            alphabet = Counter(" ".join(self.corpus.train["text"]))
            merges = []
        
        else:
            return f"Vocabulary building model ({model}) not recognized"


        vocabulary = freq.most_common()
        
        if special_tokens != None:
            vocabulary = [(token,0) for token in special_tokens] + vocabulary
        
        self.merges = merges
        self.encode = {k:i for i,(k,v) in enumerate(vocabulary)}
        self.freq = dict(vocabulary)
        self.alpha = dict(alphabet.most_common())

        self.decode = {i:k for k,i in self.encode.items()}
        
        self.len = len(vocabulary)     
        self.freq_mass = sum(self.freq.values())
        self.prob = {k:v/self.freq_mass for k,v in self.freq.items()}

        print("SLG [I]: Vocabulary built\n")
        
        if save == True:
            self.save()
            print(f"SLG [I]: Vocabulary saved to {self.path}\n")
        
        self.semiotic.syntagmatic = Syntagmatic(self.semiotic)
        self.semiotic.paradigmatic = Paradigmatic(self.semiotic)
        print("SLG [I]: Syntagmatic and Paradigmatic updated with the new vocabulary\n")

    def save(self, path = None):
        
        if path == None:
            path = self.path
            
        version_stamp = f"#version: {slg_version} - Built by `semiolog`"
        
        util.save_file([version_stamp]+self.merges,path / "merges.txt")
        util.save_file(self.encode,path / "vocab.json")
        util.save_file(self.freq,path / "freq.json")
        util.save_file(self.alpha,path / "alpha.json")
        
class nGram():
    def __init__(self, from_file = None, from_dict = None):

        if isinstance(from_dict,dict):
            self.freq = from_dict
        
        elif isfile(from_file):
            self.freq = util.load_file(from_file)
        else:
            raise Exception(f"SLG [E]: No valid dictionnary or filename provided.")

        self.keys = list(self.freq.keys())
        self.encode = {t:i for i,t in enumerate(self.keys)}
        self.decode = {i:t for i,t in enumerate(self.keys)}

        # self.prob = util.normalize_dict(self.freq)

    def extract_ngrams(text, n, count = False):

        if count:
            n_grams = Counter(zip(*[text[i:] for i in range(n)]))
        else:
            n_grams = list(zip(*[text[i:] for i in range(n)]))

        return n_grams

    def build(
        corpus=None,
        n=2,
        thres = 1,
        str_normalizer = None,
        parallel = False,
        keep_in_memory = False,
        cpu_count = 8,
        ):
        
        print("SLG [W]: This feature is not fully implemented/tested yet")
        #TODO: implement automatic loading of corpus if corpus==None
        #TODO: Dynamic cpu_count
        #TODO: implement save option

        cpu_count = 8

        def count_ngrams(corpus,n,str_normalizer):
            """
            Corpus needs to be a list of strings
            """
            corpus = " ".join(corpus)

            print(f"SLG: Counting {n}-Grams...")
            
            ngrams = Counter()
            for ng in tqdm(zip(*[corpus[i:] for i in range(n)]),total=(len(corpus)-n)):
                ngrams[ng]+=1

            return ngrams

        if str_normalizer != None:
            print(f"SLG: Normalizing Corpus...")
            corpus_norm = []
            for sent in tqdm(corpus["text"]):
                corpus_norm.append(str_normalizer(sent))

            corpus = Dataset.from_dict({"text":corpus_norm})
            del(corpus_norm)

        if parallel:
            corpus_chunks = [corpus.shard(cpu_count, n, contiguous=True, keep_in_memory = keep_in_memory) for n in range(cpu_count)]

            with Parallel(n_jobs=cpu_count, require='sharedmem') as parallel_pool:
                jobs_data = parallel_pool(delayed(count_ngrams)(chunk["text"],n,str_normalizer) for chunk in corpus_chunks)

                ngrams = reduce(operator.add,jobs_data)

        else:
        
            ngrams = count_ngrams(corpus=corpus["text"],n=n,str_normalizer=str_normalizer)

        ngrams = {"".join(tup) : freq for tup, freq in ngrams.most_common() if " " not in tup and freq>=thres}

        # if save:
        #     slg.util.save_file(ngrams,semiotic.paths.vocabulary / f"ngrams/{n}.json")

        return ngrams
    

    def __repr__(self) -> str:
        return f"nGram({self.freq})"

    # def __getitem__(self, item):
    #      return self.freq.items()[item]


# class nGram(Vocabulary):
#     def __init__(self, filename = None, special_tokens = None):

#         if filename != None:
                
#             with open(filename, "r") as f:
#                 csv_reader = csv.reader(f)
#                 voc = Counter()
#                 for line in csv_reader:
#                     voc[tuple(line[:2])] = int(line[-1])
#                 voc = dict(voc.most_common())

#             self.filename = filename
#             self.len = len(voc)
#             self.freq = voc
#             self.freq_mass = sum(voc.values())
#             self.prob = {k:v/self.freq_mass for k,v in self.freq.items()}

#             self.encode = {k:i for i,(k,v) in enumerate(voc.items())}
#             self.decode = {i:k for k,i in self.encode.items()}
#         else:
#             pass

#     def __repr__(self) -> str:
#         return f"nGram({self.freq})"

#     def __str__(self) -> str:
#         return str(self.freq)

#     def __getitem__(self, item):
#          return self.prob[item]


from .syntagmatic import Chain, Tree
from .paradigmatic import ParadigmChain
from .typing import TypeChain

class Text:
    
    def __init__(self,input_chain,semiotic,paradigms=True) -> None:

        self.chain = Chain(input_chain,semiotic)
        # semiotic.syntagmatic.tokenizer(self.chain)

        self.tree = Tree(self.chain.tree_tokens)
        
        # if semiotic.config.paradigmatic.load_pretrained != False:
        if paradigms:
            semiotic.paradigmatic.paradigmatizer(self.chain)

            self.parad_chain = ParadigmChain(self.chain)

        # semiotic.typing.typer(self.parad_chain)
        # self.type_chain = TypeChain(self.parad_chain)


from datasets import load_dataset, load_from_disk, DatasetDict, Dataset
from datasets.utils import disable_progress_bar
# disable_progress_bar()

from os.path import isfile, isdir, basename
from os import listdir
from nltk.tokenize import sent_tokenize

import socket
socket_name = socket.gethostname()
if any(name in socket_name for name in {"Gianni","vpn"}):
    from tqdm.notebook import tqdm, trange
else:
    from tqdm.auto import tqdm, trange

from .util import load_file, save_file, flatten

class Corpus:
    """
    Corpus class. It takes as argument a Huggingface's dataset object
    """
    
    def __init__(
        self,
        semiotic,
        ) -> None:
        
            self.name = semiotic.name
            self.path = semiotic.paths.corpus
            self.config = semiotic.config.corpus
        
            self.train = None
            self.test = None
            self.dev = None
            
            self.train_len = None
            self.dev_len = None
            self.test_len = None

            self.dataset = DatasetDict()
        

    def from_file(self, path = None, test_only = False):

        # TODO: loading only part of the dataset (corresponding to config.corpus.length) is not yet implemented

        if path == None:
            path = self.path
        
        if isdir(path / "dataset"):
            self.dataset = load_from_disk(path / "dataset")
            print(f"SLG [I]: Dataset loaded from disk (dataset file)")
            if not test_only:
                self.train = self.dataset["train"]
                self.train_len = self.train.num_rows

        else:
            splits = ["train","dev","test"] if test_only == False else ["dev","test"]
            filenames = [(path / f"{fn}.txt") for fn in splits]
            
            for filename in filenames:
                if not isfile(filename):
                    return print(f"SLG [W]: {filename} does not exist.\nCorpus will not be loaded from disk.\n")

            if not test_only:
                self.dataset = self.load_dataset({"train": "train.txt", "dev": "dev.txt", "test": "test.txt"})
                print(f"SLG [I]: Dataset loaded from disk (TXT files)")
                

                self.train = self.dataset["train"]
                self.train_len = self.train.num_rows

            else:
                self.dataset = self.load_dataset({"dev": "dev.txt", "test": "test.txt"})
        
        self.dev = self.dataset["dev"]
        self.test = self.dataset["test"]

        self.dev_len = self.dev.num_rows
        self.test_len = self.test.num_rows

    def load_dataset(
        self,
        dataset = None,
        original=False,
        keep_source=False
        ):

        if original:
            load_path = self.path / 'original'
        else:
            load_path = self.path

        if dataset == None:
            dataset = [fn for fn in listdir(load_path) if fn.endswith(".txt")]


        if keep_source:
            if not original:
                print("SLG [W]: 'keep_source' is True for non original corpus files. Dataset will be loaded without a source feature. If you want a source feature in your dataset, please make sure to load orignal files.")
            else:
                text = []
                source = []
                for fn in dataset:
                    # treatise_raw = txt2list(fn[:-4],self.path / 'original')
                    treatise_raw = load_file(self.path / f"original/{fn}")
                    text += treatise_raw
                    source += [fn[:-4]]*len(treatise_raw)
                
                data = DatasetDict({"train": Dataset.from_dict({"text":text, "source":source})})
        else:

            if len(dataset) == 1:
                dataset = dataset[0]

            data = load_dataset(str(load_path), data_files=dataset)

        return data


    def build(
        self,
        dataset = None,
        length = None,
        save = False,
        split_rate = None,
        split_sent = False,
        keep_source = False,
        ):
        
        if dataset == None:
            dataset = self.config.dataset

        if dataset == None:
            dataset = [fn for fn in listdir(self.path / 'original') if fn.endswith(".txt")]
            if len(dataset) == 1:
                dataset = dataset[0]

        if length == None:
            length = self.config.length
            
        if split_rate == None:
            split_rate = self.config.split_rate

        assert self.config.dataset != None or dataset != [], f"SLG [E]: No dataset defined or no txt files found in the model's folder."
        

        # if keep_source:
        #     text = []
        #     source = []
        #     for fn in dataset:
        #         treatise_raw = txt2list(fn[:-4],self.path / 'original')
        #         if split_sent:
        #             treatise_sents = []
        #             for paragraph in treatise_raw:
        #                 treatise_sents += sent_tokenize(paragraph)
        #         else:
        #             treatise_sents = treatise_raw
        #         text += treatise_sents
        #         source += [fn[:-4]]*len(treatise_sents)


        #     self.dataset = Dataset.from_dict({"text":text, "source":source})

        # else:
        if len(dataset) == 1:
            dataset = dataset[0]

        self.dataset = self.load_dataset(dataset, original=True, keep_source=keep_source)
        print(f"\nSLG [I]: Dataset loaded from the following files: {dataset}.\n")
        
        # TODO: The splitting should be generalized to handle any kind of feature potentially existing in a database
        if split_sent:
            print("Spliting original text into sentences.")

            if keep_source:
                def split(batch):
                    text_src = [(sent_tokenize(sent),src) for sent,src in zip(batch["text"],batch["source"])]
                    text = flatten([t for t,s in text_src])
                    source = flatten([[s]*len(t) for t,s in text_src])
                    split_dict = {"text": text, "source": source}
                    return split_dict
                    
                for column in self.dataset.column_names:
                    self.dataset[f"{column}"] = self.dataset[f"{column}"].map(split, remove_columns=["text","source"], batched=True)

            else:
                for column in self.dataset.column_names:
                    self.dataset[f"{column}"] = self.dataset[f"{column}"].map(lambda batch: {"text": flatten([sent_tokenize(sent) for sent in batch["text"]])}, remove_columns=["text"], batched=True)

        if "test" in self.dataset:
            if length != None:
                split_lengths = tuple([int(length*r) for r in split_rate])
            print("This feature has not been tested yet. Please check")

            #TODO: Check the entire "if". Use "train_test_split" instead 

            self.train = self.dataset["train"][:split_lengths[0]]
            
            if "dev" or "validation" in self.dataset:
                
                self.dev = self.dataset.get("dev",self.dataset["validation"])[:split_lengths[1]]
                self.test = self.dataset["test"][:split_lengths[2]]
                
            else:
                
                split_test = self.dataset["test"][:split_lengths[1]+split_lengths[2]].train_test_split(split_rate[0]*split_rate[1])

                self.dev = split_test["train"]
                self.test = split_test["test"]
            
            self.dataset = DatasetDict({"train": self.train, "dev": self.dev, "test": self.test})
                    
        
        else:

            dataset_train = self.dataset["train"].train_test_split(test_size=sum(split_rate[1:]))

            dataset_test = dataset_train["test"].train_test_split(split_rate[1]/sum(split_rate[1:]))

            self.dataset = DatasetDict({"train": dataset_train["train"], "dev": dataset_test["train"], "test": dataset_test["test"]})
            
            self.sentences = self.dataset["train"]["text"]

            self.train = self.dataset["train"]
            self.dev = self.dataset["dev"]
            self.test = self.dataset["test"]
            
        self.train_len = self.train.num_rows
        self.dev_len = self.dev.num_rows
        self.test_len = self.test.num_rows
        
        print("Corpus built")
        if save:
            if keep_source:
                self.save(dataset_format=True)
                print(f"Corpus saved as dataset file to {self.path / 'dataset'}")

            else:
                self.save()
                print(f"Corpus saved as TXT files to {self.path}")
        
        
    def save(self, path = None, dataset_format=False):
        
        if path == None:
            path = self.path

        if dataset_format:
            self.dataset.save_to_disk(path / "dataset")
        else:
            save_file(self.train["text"], path / "train.txt")
            save_file(self.dev["text"], path / "dev.txt")
            save_file(self.test["text"], path / "test.txt")


import string
import unicodedata
import re

# punctuation = '...—•…–’‘-·[]⁄′¿"‐―'
punctuation = '...—•…–’‘·⁄′¿"‐―'

class Normalizer:
    """
    """
    def __init__(self) -> None:
        pass

    def normalize(self, input_string:str):
        pass

class disable:
    """
    Disable this step. It returns the input as output #TODO: maybe with the correct type for the pipeline
    """
    def __init__(self) -> None:
        pass

    def normalize(self, input_string:str):
        return input_string

class Sequence(Normalizer):
    def __init__(self,sequence:list) -> None:
        self.sequence = [eval(normalizer_sequence)() for normalizer_sequence in sequence]
    
    def __repr__(self) -> str:
        return f"Sequence({self.sequence})"

    def normalize(self, input_string: str):

        output = input_string
        for normalizer_component in self.sequence:
            output = normalizer_component.normalize(output)
        return output


class NFKC(Normalizer):
    """
    NFKC Normalizer using unicodedata module

    Args:
        Normalizer ([type]): [description]

    Returns:
        [type]: [description]
    """
    def __init__(self) -> None:
        super().__init__()
    
    def normalize(self, input_string: str):
        return unicodedata.normalize("NFKC", input_string)

class NFKD(Normalizer):
    """
    NFKD Normalizer using unicodedata module

    Args:
        Normalizer ([type]): [description]

    Returns:
        [type]: [description]
    """
    def __init__(self) -> None:
        super().__init__()
    
    def normalize(self, input_string: str):
        return unicodedata.normalize("NFKD", input_string)

class Lowercase(Normalizer):
    """
    """
    def __init__(self) -> None:
        super().__init__()
    
    def normalize(self, input_string: str):
        return input_string.lower()


class StripPunctuation(Normalizer):
    """
    """
    def __init__(self) -> None:
        super().__init__()
    
    def normalize(self, input_string: str):
        return input_string.translate(str.maketrans('', '', string.punctuation+punctuation))

class StripWhitespaces(Normalizer):
    """
    """
    def __init__(self) -> None:
        super().__init__()
    
    def normalize(self, input_string: str):
        return input_string.translate(str.maketrans('', '', string.whitespace))

class Foreign(Normalizer):
    """
    Replace Chinese, Japanese, Korean, Arabic, Cyrillic, Armenian, Hebrew, Syriac, Devanagari, Bengali characters
    """
    def __init__(self) -> None:
        super().__init__()
    
    def normalize(self, input_string: str):
        
        chars = re.compile(
            r"[\uac00-\ud7a3]|[\u4e00-\u9FFF]|[\u3040-\u30ff]|[\u0400-\u04FF]|[\u0500-\u052F]|[\u0530-\u058F]|[\u0590-\u05FF]|[\u0600-\u06FF]|[\u0700-\u074F]|[\u0750-\u077F]|[\u0980-\u09FF]|[\u0900-\u097F]"
            )
        
        
        return chars.sub("□", input_string)


# GENERAL FUNCTIONS #
#####################

# Load Modules

import os, sys
import math
import concurrent.futures
from joblib import Parallel, delayed
import csv
import json
import itertools
import functools
import ast
from collections import Counter
import numpy as np
from scipy.sparse import csr_matrix
import time
import pandas as pd
import matplotlib.pyplot as plt
import plotly.graph_objects as go

import socket
socket_name = socket.gethostname()
if any(name in socket_name for name in {"Gianni","vpn"}):
    from tqdm.notebook import tqdm, trange
else:
    from tqdm.auto import tqdm, trange

# Definitions of Functions

def chunk_size(total_n, chunks_n):
    return math.ceil(total_n / chunks_n)

def partition(lst, n_of_parts):
    """Partitions list in n parts."""
    if not isinstance(lst,list):
        lst = list(lst)
    chunk_size = math.ceil(len(lst) / n_of_parts)
    for i in range(0, len(lst), chunk_size):
        yield lst[i : i + chunk_size]

# head and chunks taken from https://stackoverflow.com/questions/24527006/split-a-generator-into-chunks-without-pre-walking-it

def head(iterable, max=10):
    first = next(iterable)      # raise exception when depleted
    def head_inner():
        yield first             # yield the extracted first element
        for cnt, el in enumerate(iterable):
            yield el
            if cnt + 1 >= max:  # cnt + 1 to include first
                break
    return head_inner()

def chunks(iterable, size=10):
    iterator = iter(iterable)
    for first in iterator:
        yield itertools.chain([first], itertools.islice(iterator, size - 1))
        
def multiprocessing(func, args, cores=None):
    if cores == None:
        cores = os.cpu_count()

    result = Parallel(n_jobs=cores)(delayed(func)(i) for i in args)
    return result
    
    # with concurrent.futures.ProcessPoolExecutor(cores) as executor:
    #     result = executor.map(func, args, chunksize=chunksize)
    # return list(result)

def multithreading(func, args, chunksize=1,cores=None):
    with concurrent.futures.ThreadPoolExecutor(cores) as executor:
        result = executor.map(func, args, chunksize=chunksize)
    return list(result)


# Adapted Dans Shiebler: from http://danshiebler.com/2016-09-14-parallel-progress-bar/

def multiprocessing_tqdm(function, array, cores=None, desc = None):
    """
        A parallel version of the map function with a progress bar. 

        Args:
            array (array-like): An array to iterate over.
            function (function): A python function to apply to the elements of array
            n_jobs (int, default=16): The number of cores to use
            use_kwargs (boolean, default=False): Whether to consider the elements of array as dictionaries of 
                keyword arguments to function 
            front_num (int, default=3): The number of iterations to run serially before kicking off the parallel job. 
                Useful for catching bugs
        Returns:
            [function(array[0]), function(array[1]), ...]
    """
    #We run the first few iterations serially to catch bugs
    # if front_num > 0:
    #     front = [function(**a) if use_kwargs else function(a) for a in array[:front_num]]
    # #If we set n_jobs to 1, just run a list comprehension. This is useful for benchmarking and debugging.
    # if n_jobs==1:
    #     return front + [function(**a) if use_kwargs else function(a) for a in tqdm(array[front_num:])]
    #Assemble the workers
    with concurrent.futures.ProcessPoolExecutor(max_workers=cores) as pool:
        #Pass the elements of array into function

        futures = [pool.submit(function, a) for a in array]
        kwargs = {
            'total': len(futures),
            'unit': 'it',
            'unit_scale': True,
            'leave': True
        }
        #Print out the progress as tasks complete
        for f in tqdm(concurrent.futures.as_completed(futures), **kwargs, desc = desc):
            pass
    out = []
    #Get the results from the futures. 
    for future in futures:
        try:
            out.append(future.result())
        except Exception as e:
            out.append(e)
    return out

def save_file(data, filename):
    
    if os.path.exists(filename):
        raise Exception(f"SLG [E]: {filename} already exist [TODO: implement overwrite]")

    fn_dir = os.path.dirname(filename)
    fn_extension = os.path.splitext(filename)[-1]

    if not os.path.isdir(fn_dir):
        os.makedirs(fn_dir)

    if fn_extension == ".json":
        with open(filename, 'w', encoding='utf-8') as json_file:
            json.dump(data, json_file,indent=4, ensure_ascii=False)

    elif fn_extension == ".txt":
        with open(filename, "w", encoding='utf-8') as txt_file:
            for element in data:
                txt_file.write(element+"\n")
    
    else:
        raise Exception(f"SLG [E]: File extension {fn_extension} not recognised")



def load_file(filename):
    
    if not os.path.isfile(filename):
        raise Exception(f"SLG [E]: {filename} does not exist")
    
    fn_extension = os.path.splitext(filename)[-1]

    if fn_extension == ".json":
        with open(filename) as json_file:
            data = json.load(json_file)

    elif fn_extension == ".txt":
        with open(filename) as txt_file:
            data = [line.strip() for line in txt_file]

    else:
        raise Exception(f"SLG [E]: File extension {fn_extension} not recognised")

    return data

def dict2csv(input: dict, filename: str, path: str):
    if not os.path.isdir(path):
        os.makedirs(path)
    with open(f"{path}{filename}.csv", "w") as nf:
        for key in input.keys():
            nf.write(f"{key},{input[key]}\n")

def dict2json(input: dict, filename: str, path: str):
    if not os.path.isdir(path):
        os.makedirs(path)
    with open(f"{path}/{filename}.json", 'w',encoding='utf-8') as json_file:
        json.dump(input, json_file,indent=4, ensure_ascii=False)



def json2dict(filename, path):
    with open(f"{path}/{filename}.json") as json_file:
        data = json.load(json_file)
    return data

def str2txt(input: str, filename: str, path: str):
    if not os.path.isdir(path):
        os.makedirs(path)
    with open(f"{path}{filename}.txt", "w") as nf:
        nf.write(input)

# def list2txt(input: list, filename: str = "text_list", path = None):
#     if not os.path.isdir(path):
#         os.makedirs(path)
#     with open(f"{path}/{filename}.txt", "w") as nf:
#         for element in input:
#             nf.write(element+"\n")
            
# def txt2list(filename, path):
#     with open(f"{path}/{filename}.txt") as txt_file:
#         txt_list = [line.strip() for line in txt_file]
#     return txt_list

def subsequences(sequence, n: int):
    list_of_tuples = [tuple(sequence[i:i+n]) for i in range(len(sequence)-n+1)]
    return list_of_tuples

def powerset(iterable):
    "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    s = list(iterable)
    return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s)+1))

def list2csv(input: list, filename: str, directory: str):
    if not os.path.isdir(directory):
        os.makedirs(directory)
    with open(f"/{directory}{filename}.csv", "w", newline="") as nf:
        wr = csv.writer(nf, quoting=csv.QUOTE_ALL)
        wr.writerows(input)

def csv2list(file: str, directory: str, n_start=None, n_end=None):
    if not os.path.isdir(directory):
        os.makedirs(directory)
    with open(f"{directory}{file}.csv", "r") as f:
        csv_reader = csv.reader(f)
        my_list = [
            (line[0], ast.literal_eval(line[1]), ast.literal_eval(line[2]))
            for line in take(n_end, csv_reader)
        ]
    return my_list[n_start:n_end]

# TODO: Really needed here?
def if_none_disable(x):
    if x == None:
        x = "disable"
    return x


def take(n, iterable):
    """Return first n items of the iterable as a list"""
    return list(itertools.islice(iterable, n))

def marginalize(joint_dist:dict,side="left"):
    # print(f"Marginalizing to the {side}")
    marginal = Counter()
    for (l,r),v in joint_dist.items():
        marginal[r if side == "right" else l] += v
    return dict(marginal.most_common())

def normalize_dict(dictionnary: dict, norm_factor = None):
    if norm_factor == None:
        norm = 1/sum(dictionnary.values())
    else:
        norm = norm_factor
    prob_dict = {k: v * norm for k, v in dictionnary.items()}
    return prob_dict

def pmi(
    matrix,
    alpha=.75,
    type_pmi="sppmi",
    ):
    """
    """
    # Taken from Kaggle, and modified

    if type_pmi == "nopmi":
        return matrix

    if type_pmi not in {"pmi","npmi","ppmi","nppmi","spmi","sppmi","ssppmi","nopmi"}:
        pmi = "sppmi"
        print("Unknown type of PMI. Continuing with smoothed positive PMI (SPPMI)")

    num_skipgrams = matrix.sum()
    matrix_items = {(i, j): matrix[i,j] for i, j in zip(*matrix.nonzero())}
    # assert(sum(matrix_items.values()) == num_skipgrams)


    # for creating sparse matrices
    row_indxs = []
    col_indxs = []

    pmi_dat_values = []    # pointwise mutual information

    # reusable quantities
    sum_over_contexts = np.array(matrix.sum(axis=1)).flatten()
    sum_over_terms = np.array(matrix.sum(axis=0)).flatten()

    # smoothing
    sum_over_terms_alpha = sum_over_terms**alpha
    nca_denom = np.sum(sum_over_terms_alpha)

    for (tok_terms, tok_context), sg_count in matrix_items.items():
        # here we have the following correspondance with Levy, Goldberg, Dagan
        #========================================================================
        #   num_skipgrams = |D|
        #   nwc = sg_count = #(w,c)
        #   Pwc = nwc / num_skipgrams = #(w,c) / |D|
        #   nw = sum_over_cols[tok_ingredient]    = sum_over_contexts[tok_ingredient] = #(w)
        #   Pw = nw / num_skipgrams = #(w) / |D|
        #   nc = sum_over_rows[tok_context] = sum_over_ingredients[tok_context] = #(c)
        #   Pc = nc / num_skipgrams = #(c) / |D|
        #
        #   nca = sum_over_rows[tok_context]^alpha = sum_over_ingredients[tok_context]^alpha = #(c)^alpha
        #   nca_denom = sum_{tok_content}( sum_over_ingredients[tok_content]^alpha )

        nwc = sg_count
        Pwc = nwc / num_skipgrams
        nw = sum_over_contexts[tok_terms]
        Pw = nw / num_skipgrams
        
        # note 
        # pmi = log {#(w,c) |D| / [#(w) #(c)]} 
        #     = log {nwc * num_skipgrams / [nw nc]}
        #     = log {P(w,c) / [P(w) P(c)]} 
        #     = log {Pwc / [Pw Pc]}

        if type_pmi == "pmi":
            nc = sum_over_terms[tok_context]
            Pc = nc / num_skipgrams
            pmi = np.log(Pwc/(Pw*Pc))

        elif type_pmi == "npmi":
            nc = sum_over_terms[tok_context]
            Pc = nc / num_skipgrams
            pmi = np.log(Pwc/(Pw*Pc))/-np.log(Pwc)
        
        elif type_pmi == "ppmi":
            nc = sum_over_terms[tok_context]
            Pc = nc / num_skipgrams
            pmi = max(np.log(Pwc/(Pw*Pc)), 0)

        elif type_pmi == "nppmi":
            nc = sum_over_terms[tok_context]
            Pc = nc / num_skipgrams
            pmi = max(np.log(Pwc/(Pw*Pc))/-np.log(Pwc),0)

        elif type_pmi == "spmi":
            nca = sum_over_terms_alpha[tok_context]
            Pca = nca / nca_denom
            pmi = np.log(Pwc/(Pw*Pca))
        elif type_pmi == "sppmi":
            nca = sum_over_terms_alpha[tok_context]
            Pca = nca / nca_denom  
            pmi = max(np.log(Pwc/(Pw*Pca)), 0)
        elif type_pmi == "ssppmi":
            nc = sum_over_terms[tok_context]
            Pc = nc / num_skipgrams
            pmi = max(np.log(Pwc/(Pw*Pc*alpha)), 0)
        
        row_indxs.append(tok_terms)
        col_indxs.append(tok_context)
        pmi_dat_values.append(pmi)
            
    pmi_mat = csr_matrix((pmi_dat_values, (row_indxs, col_indxs)),shape=matrix.shape)

    print('Done')
    return pmi_mat



def mm_no_modif(contexts, orthogonals, term):

    if isinstance(next(iter(orthogonals)),str):
        return [orthogonals.get(term+context,0) for context in contexts]
    elif isinstance(next(iter(orthogonals)),tuple):
        return [orthogonals.get((term, context),0) for context in contexts]
    else:
        raise Exception(f"SLG [E]: Orthogonal dictionnary keys have the wrong type ({type(type(next(iter(orthogonals))))}). Types accepted: str and tuple ")

def matrix_maker(
    terms,
    contexts,
    orthogonals,
    measure = mm_no_modif):
    results = multithreading(functools.partial(measure, contexts, orthogonals), terms)
    return results

def build_term_context_matrix(
    terms,
    contexts,
    orthogonals,
    normalizeQ = False):
    print("Building oR Matrix...")
    start = time.perf_counter()
    if normalizeQ:
        orthogonals = normalize_dict(orthogonals)
    matrix = csr_matrix(matrix_maker(terms,contexts,orthogonals))
    finish = time.perf_counter()
    print(f"Term-Context Matrix built in {round(finish-start,2)} secs.\n")
    return matrix

def build_pmi_matrix(
    term_context_matrix,
    type = "pmi",
    alpha = .75,
    normalizeQ = False,
    ):

    print("Computing PMI Matrix...")
    print(f"Type: {type}")
    if "s" in type:
        print(f"Smoothing (alpha): {alpha}")
    start = time.perf_counter()
    pmi_matrix = pmi(term_context_matrix,alpha=alpha,type_pmi=type)
    finish = time.perf_counter()
    if normalizeQ:
        print("Normalizing Matrix")
        pmi_matrix = (1/(pmi_matrix.sum()))*pmi_matrix
    print(f"PMI Matrix built in {round(finish-start,2)} secs.")
    print("Done\n")
    return pmi_matrix



def parallel_processes(processes:list):
    """
    Warning: this produces global variables "parallel_process_n" which are deleted at the end of the computation.
    """
    st = time.perf_counter()
    np = len(processes)
    print(f"Computing {np} processes in parallel...")
    executor = concurrent.futures.ProcessPoolExecutor(max_workers=np)
    for n, process in zip(range(np),processes):
        variable_name = f"parallel_process_{n}"
        globals()[variable_name] = executor.submit(*process)
    for n in range(np):
        variable_name = f"parallel_process_{n}"
        globals()[variable_name] = eval(variable_name).result()
    fin = time.perf_counter()
    print(f"{np} processes computed in {round(fin-st,2)} secs.\n")
    result = [eval(f"parallel_process_{n}") for n in range(np)]
    for n in range(np):
        del globals()[f"parallel_process_{n}"]
    return result

def flatten(list_of_lists):
    return [item for sublist in list_of_lists for item in sublist]

def delete_duplicates(x):
    return list(dict.fromkeys(x))

def clear_df(df):
    mask = df.applymap(lambda x: x is None)
    cols = df.columns[(mask).any()]
    for col in df[cols]:
        df.loc[mask[col], col] = ''
    return df

def df(list_of_list, keys = None):
    raw_df = pd.DataFrame(list_of_list,index=keys)
    return clear_df(raw_df.T)

# Draft functions

# def csv2list_raw(file: str, directory: str, n_start=None, n_end=None):
#     if not os.path.isdir(directory):
#         os.makedirs(directory)
#     with open(f"{directory}{file}.csv", "r") as f:
#         csv_reader = csv.reader(f)
#         my_list = [
#             [ast.literal_eval(item) for item in line]
#             for line in take(n_end, csv_reader)
#         ]
#     return my_list[n_start:n_end]

def csv2list_raw(file: str, directory: str, n_start=None, n_end=None,progress=False):
    if not os.path.isdir(directory):
        os.makedirs(directory)
    with open(f"{directory}{file}.csv", "r") as f:
        csv_reader = csv.reader(f)
        my_list = []
        if progress:
            csv_reader_it = tqdm(take(n_end, csv_reader),desc = f"Loading: {file}")
        else:
            csv_reader_it = take(n_end, csv_reader)
        for line in csv_reader_it:
            list_line = [ast.literal_eval(item) for item in line]
            my_list.append(list_line)
    return my_list[n_start:n_end]

def plot_scatter_line(
    x,
    y,
    trace_name = None,
    title=None,
    xaxis_title=None,
    yaxis_title=None,
    line_shape='spline',
    add_trace = None
    ):
    fig = go.Figure()
    fig.add_trace(go.Scatter(x=x, y=y, name = trace_name, line_shape=line_shape, mode='lines+markers'))
    
    if add_trace != None:
        if isinstance(add_trace,tuple):
            add_trace = [add_trace]
        for trace in add_trace:
            fig.add_trace(go.Scatter(x=trace[0], y=trace[1], name = trace[2], line_shape=line_shape, mode='lines+markers')) # TODO: rewrite with **kwargs


    fig.update_layout(
        title=title,
        xaxis_title=xaxis_title,
        yaxis_title=xaxis_title,
        
        )
    return fig

def scatter_plot(data:list,labels=None,legends=None,size=(15,10)):
    fig, ax = plt.subplots(figsize=size)  # Create a figure and an axes.
    for dataset in data:
        ax.scatter(range(len(dataset)) if labels==None else labels,dataset,label="Data")
    ax.set_xlabel('x')  # Add an x-label to the axes.
    ax.set_ylabel('f(x)')  # Add a y-label to the axes.
    ax.set_title("Title")  # Add a title to the axes.
    ax.legend(range(1,len(data)+1) if legends==None else legends)  # Add a legend.
    return ax

profiler = """
from pyinstrument import Profiler
import sys
def profiler(process):
    profile = Profiler()

    profile.start()
    
    try:
        exec(process)
    except:
        print(f"{sys.exc_info()}")

    profile.stop()

    return print(profile.output_text(unicode=True, color=True))"""


# value and index of max of matrix, in case needed
# np.array(np.unravel_index(np.argsort(voc_matrix.toarray().flatten(), axis=0)[-10:], voc_matrix.shape)).T[::-1]



from scipy.stats import entropy
import numpy as np
import tensorflow as tf
import string
from ..syntagmatic.tokenizer import normalizers
from collections import Counter, defaultdict
from operator import itemgetter
from math import log

from ..util import df as slg_df

def tolerance_principle(n):
    return n/log(n)

# TODO: Consider different subtypings in the determination of tolerance
# TODO: Consider also prefixes and infixes

def productive_suffix(parad_keys):
    max_len = max([len(term) for term in parad_keys])

    parad_len = len(parad_keys)

    # TODO This Threshold is set ad hoc and as a quick fix. Think about a better way to avoid division by 0 in tolerance_principle when parad_len = 1. And also to avoid considering productive parads of len 1 or 0.
    if parad_len <2:
        prod_thres = 2
    else:
        prod_thres = parad_len - tolerance_principle(parad_len) #Tolerance principle (Yang, 2016)

    # l=1

    sub_parad = parad_keys
    sub_parad_collector = {key:key for key in parad_keys}

    for suffix_len in range(1,max_len):
        suffix_counter = Counter()
        suffix_dict = defaultdict()
        for term in parad_keys:
            suffix = term[-suffix_len:]
            suffix_counter[suffix]+=1
            suffix_dict[suffix] = suffix_dict.get(suffix,[])+[term]

        if  max(suffix_counter.values()) >= prod_thres:
            suffix = []
            suffix_freq = []
            for s,f in suffix_counter.most_common():
                suffix.append(s)
                suffix_freq.append(f)

            sub_parad = suffix_dict[suffix[0]]
            # TODO This Threshold is set ad hoc and as a quick fix. Think about a better way to avoid division by 0 in tolerance_principle when parad_len = 1. And also to avoid considering productive parads of len 1 or 0.
            if len(sub_parad) <2:
                prod_thres = 2
            else:
                prod_thres = len(sub_parad)-tolerance_principle(len(sub_parad)) # len(sub_parad)/2

            for term in sub_parad:
                sub_parad_collector[term] = (term[:-suffix_len],suffix[0])
            
        else:
            break
    if any([isinstance(term,tuple) for term in sub_parad_collector.values()]):
        return list(sub_parad_collector.values())
    else:
        return None

class Paradigm:
    def __init__(
        self,
        ids,
        values,
        non_zeroes,
        decoder,
        semiotic,
        cum_thres=.5,
        vocab_thres = 500
        ) -> None:

        self.len = non_zeroes.numpy()
        self.ids = ids.numpy()
        self.keys = decoder(ids.numpy()[:self.len]).split()
        self.values = values.numpy()
        query = itemgetter(*self.keys)
        self.probs = query(semiotic.vocab.prob)
        if isinstance(self.probs,float): # if len of keys is 1, then query outputs a float instead of a tuple
            self.probs = tuple([self.probs])
        self.mass = sum(self.probs)

        self.entropy = entropy(self.values)
        self.mass_entropy = entropy(self.probs)
        # self.cumsum = np.cumsum(self.values)

        key_thres = set(list(semiotic.vocab.prob.keys())[:vocab_thres])
        self.top_freq = [key for key in self.keys if key in key_thres]

        self.productivity = productive_suffix(self.keys)


    def __repr__(self) -> str:
        return str(self.keys)

class Paradigmatizer:
    
    def __init__(self, model, bert_tokenizer, decoder) -> None:
        self.model = model
        self.bert_tokenizer = bert_tokenizer
        self.decoder = decoder
        
    def __call__(self,chain,n_token=None):
        """
        If n_token == None, then it computes paradigms for all tokens and adds them as attributes to tokens in chain. If a token is provided, then returns the paradigm for that token.
        """
        
        # TODO: inputs longer than self.model.config.max_position_embeddings yield an error. The workaround here is to truncate the input with [:,:self.model.config.max_position_embeddings], but then not all tokens receive their paradigm. A more consistent solution is needed.

        if isinstance(n_token,int):
            if n_token>=min(chain.len, self.model.config.max_position_embeddings):
                raise Exception(f"The toke number should be within the range of the chain length (< {min(chain.len, self.model.config.max_position_embeddings)})")
            sent_mask = chain.mask(n_token)
            input = self.bert_tokenizer(sent_mask)
            outputs = self.model(input["input_ids"][:,:self.model.config.max_position_embeddings])
            parad_logits = outputs.logits[0, n_token]
            logits_positif = tf.nn.relu(parad_logits)
            probs, norms = tf.linalg.normalize(logits_positif, ord=1)

            non_zeroes = tf.math.count_nonzero(probs)

            parad_data = tf.math.top_k(probs, k = non_zeroes.numpy(), sorted=True)

            parad = Paradigm(
                parad_data.indices,
                parad_data.values,
                non_zeroes,self.decoder, chain.semiotic)

            return parad

        else:
            sent_mask = [chain.mask(n) for n in range(chain.len)]
            input = self.bert_tokenizer(sent_mask)
            outputs = self.model(input["input_ids"][:,:self.model.config.max_position_embeddings])
            parad_logits = tf.gather_nd(outputs.logits, indices=[[i,i] for i in range(chain.len)])
            logits_positif = tf.nn.relu(parad_logits)
            probs, norms = tf.linalg.normalize(logits_positif, ord=1, axis=1)
            
            # Maybe its cheaper to compute top k for k = non_zero, which would differ from row to row, and hence not yeld a tensor. To be tested.

            non_zeroes = tf.math.count_nonzero(probs,axis=1)
            max_non_zeroes = max(non_zeroes).numpy()

            parad_data = tf.math.top_k(probs, k = max_non_zeroes, sorted=True)

            parads = [Paradigm(ids,values,nonzeroes,self.decoder, chain.semiotic) for ids,values,nonzeroes in zip(parad_data.indices,parad_data.values,non_zeroes)]

            for token,parad in zip(chain,parads):
                token.paradigm = parad


class ParadigmChain:

    def __init__(self,chain) -> None:
        self.semiotic = chain.semiotic
        self.len = chain.len
        # self.probs = chain.probs
        self.paradigms = [token.paradigm for token in chain.tokens]
        self.keys = [token.paradigm.keys for token in chain.tokens]
        # self.keys_t = [token.paradigm.keys_t for token in chain.tokens]
        # self.keys_t_soft = [token.paradigm.keys_t_soft for token in chain.tokens]

        self.labels = [token.label for token in chain.tokens]
        self.spans = [token.span for token in chain.tokens]
        self.indexes = [f"{token.label}_{token.position}" for token in chain.tokens]
    
    def __getitem__(self, index:str):
        return self.paradigms[index]
    
    def df(self, size = 20):
        return slg_df([p[:size] for p in self.keys], self.indexes)


from .tokenizer import Tokenizer
from .chain import Chain
from .tree import Tree

class Syntagmatic:
    def __init__(self,semiotic) -> None:
        
        self.config = semiotic.config.syntagmatic
        # self.tokenizer = Tokenizer(self.config)
    
    @property
    def tokenizer(self):
        return Tokenizer(self.config)
        


import inspect
import sys
from os.path import isfile

from .util import dict2json, json2dict

# TODO: Solve version as global variable
slg_version = "0.2.3"

class Config:
    
    def __init__(self, semiotic) -> None:
        """
        Automatically loads all classes below as section attributes of the class Config
        """
        
        self.sections = [s for s,repr in inspect.getmembers(sys.modules[__name__], inspect.isclass) if s not in ("Config","Section")]
        
        for section in self.sections:
            exec(
                f"self.{section.lower()} = {section}(semiotic)"
            )

        self.path = semiotic.paths.semiotic

    def __repr__(self) -> str:
        return str(self.__dict__)

    def save(self):
        """
        Saves the configuration asa JSON file in the root of the current model. Paths are not saved
        """
        config_dict = {k: v.__dict__ for k,v in self.__dict__.items() if k in {s.lower() for s in self.sections}}

        dict2json(config_dict,"config", self.path)
    
    def from_file(self,path = None):
        if path == None:
            path = self.path
        
        filename = str(path / "config.json")
        if not isfile(filename):
            return print(f"SLG [W]: {filename} does not exist.\Config will not be loaded from disk.\n")
            
        config_dict = json2dict("config",path)
        
        for section in config_dict:
            for key in config_dict[section]:
                setattr(eval(f"self.{section}"), key, config_dict[section][key])


class Section:
    
    def __init__(self, semiotic) -> None:
        pass
    
    def __repr__(self) -> str:
        return str(self.__dict__)
    
    
class System(Section):
    
    def __init__(self, semiotic) -> None:
        self.slg_version = slg_version
        self.cpu_count = None
        self.multiple_gpus = False


class General(Section):
    """
    Possible tensor implementations: 'tf' (TensorFlow), 'pt' (PyTorch), 'np' (numpy)
    """
    def __init__(self, semiotic) -> None:
        self.name = semiotic.name
        self.dataset = None
        self.tensor_implementation = "tf"

class Corpus(Section):
    
    def __init__(self, semiotic) -> None:
        self.dataset = None
        self.split_rate = (.9,.05,.05)
        self.length = None
        
class Vocabulary(Section):
    
    def __init__(self, semiotic) -> None:
        """
        Possible models: "SLG", "SLG_WL"
        
        By default, the special tokens are:
        [PAD], [UNK], [CLS], [SEP], [MASK]
        """
        self.model = "SLG"
        self.size = None
        self.special_tokens = [
            "[PAD]",
            "[UNK]",
            "[CLS]",
            "[SEP]",
            "[MASK]"
            ]
        self.unk_token="[UNK]"
        self.pad_token="[PAD]"
        self.cls_token="[CLS]"
        self.sep_token="[SEP]"
        self.mask_token="[MASK]"
        
class Syntagmatic(Section):
    """
    Possible Normalizers: "NFKD","Lowercase","StripPunctuation","StripWhitespaces"
    
    Possible Processors: "SequenceSLG" "TreeSLG", "StripWhitespaces"m
    """
    def __init__(self, semiotic) -> None:
        self.from_file = False
        self.normalizer = None
        self.pre_tokenizer = None
        self.processor = "WordLevel"
        self.post_processor = None
        self.model_max_length = None
        self.n_sents = None
        self.input_tokenize = {
            "batched": True,
            "batch_size": 512,
            "remove_columns": ["text"],
        }
        self.save_tokenized = True


class Paradigmatic(Section):
    """
    'learning_rate' and 'weight_decay'
    """
    def __init__(self, semiotic) -> None:

        self.load_pretrained = False

        self.huggingface_pretrained = None

        self.load_tokenized = True
        self.n_sents = None

        self.model_config = {
            "hidden_size": 768,
            "num_hidden_layers": 12,
            "num_attention_heads": 12,
            "intermediate_size": 3072,
            "hidden_act": "gelu",
            "hidden_dropout_prob": 0.1,
            "attention_probs_dropout_prob": 0.1,
            "max_position_embeddings": 512,
            "type_vocab_size": 2,
            "initializer_range": 0.02,
            "layer_norm_eps": 1e-12,
            "pad_token_id": 0,
            "position_embedding_type": "absolute",
            "use_cache": True,
            "classifier_dropout": None,

            "learning_rate": 2e-5,
            "weight_decay": 0.01
        }
        
        self.optimizer = {
            "optimizer": "AdamWeightDecay",

            "learning_rate": 2e-5,
            "weight_decay": 0.01
        }

        self.mask_probability = 0.15

        self.input_sets = {
            "shuffle": True,
            "batch_size": 16
        }

        self.training_epochs = 2

        self.save = True

        self.model = None
        self.top_k = None
        self.exclude_punctuation = None
        self.cumulative_sum_threshold = None


class Evaluation(Section):
    def __init__(self, semiotic) -> None:
        self.ud_model = None
        self.cp_model = None


# FILE LOCATIONS

from pathlib import Path
from os import path

# TODO: Create all the directories referred to here

class Paths:
    def __init__(self, name) -> None:
        # TODO: configure online repository for models, with automatic download
        # self.cache = Path(path.expanduser("~/.cache/semiolog"))
        # self.online_models = "https://polybox.ethz.ch/index.php/s/bsbycrxtbeKvgEE"
        # self.models = self.cache / "models"
        
        self.models = Path("././models")
        self.semiotic = self.models / name
        self.corpus = self.models / name / "corpus"
        self.vocabulary = self.models / name / "vocabulary"
        self.paradigms = self.models / name / "paradigms"
        self.syntagmas = self.models / name / "syntagmas"
        
    def __repr__(self) -> str:
        return str(self.__dict__)


from ..util import flatten
import networkx as nx
import graphviz as gv
from ipywidgets import interactive, interact
from IPython.display import display

class Tree:
    def __init__(self, tokens):
        self.edges = [((token.head.label,token.head.span),(token.label,token.span)) for token in tokens]
        self.nodes = set(flatten(self.edges))
        self.root = [node for node in self.nodes if node not in [r for l, r in self.edges]][
            0
        ]
        self.leaves = {node for node in self.nodes if node not in [l for l, r in self.edges]}
        self.non_terminals = set(self.nodes) - self.leaves

    def graph(self):
        seg_tree = nx.DiGraph()
        seg_tree.add_edges_from(self.edges)

        return seg_tree

    
    def plot(self, red=None, grey=None, fname=f"seg_graph"):

        seg_tree = self.graph()

        seg_tree_graph = gv.Digraph(name=fname)

        tree_nodes_list = [(str(i), l) for l, i in list(seg_tree.nodes)]
        for node in tree_nodes_list:
            seg_tree_graph.node(
                *node, color="white", fontsize="20", fontname="RobotoMono-Thin"
            )  # style="filled", color="grey")
        seg_tree_graph.attr("edge", color="slategrey")
        seg_tree_graph.edges([(str(p[1]), str(c[1])) for p, c in list(seg_tree.edges)])

        if grey == None:
            pass
        else:
            for node in [(str(i), l) for l, i in grey]:
                if node in tree_nodes_list:
                    seg_tree_graph.node(
                        *node,
                        style="dashed",
                        color="grey",
                        fontsize="20",
                        fontname="RobotoMono-Thin",
                    )

        if red == None:
            pass
        else:
            for node in [(str(i), l) for l, i in red]:
                if node in tree_nodes_list:
                    seg_tree_graph.node(
                        *node, color="red", fontsize="20", fontname="RobotoMono-Thin"
                    )
        
        @interact
        def inter():
            return seg_tree_graph

        return inter


from ..functive import Functive


class ChainIterator:
    ''' Iterator class '''
    def __init__(self, chain):
        self.iterable = chain.tokens
        self.len = chain.len
        # member variable to keep track of current index
        self.index = 0
    def __next__(self):
        if self.index < self.len:
            result = self.iterable[self.index]
            self.index +=1
            return result
        raise StopIteration

class Chain:
    def __init__(self, input_chain: str,semiotic):
        
        self.semiotic = semiotic
        self.input = input_chain
        self.split = input_chain.split()
        if self.semiotic.syntagmatic.tokenizer.normalizer != None:
            self.split_norm = [self.semiotic.syntagmatic.tokenizer.normalizer.normalize_str(s) for s in self.split]
            self.split_norm = [t for t in self.split_norm if t!='']
        else:
             self.split_norm = self.split

        hf_output = semiotic.syntagmatic.tokenizer.encode(self.input)

        #TODO: Using HF tokenizers, the tokens for trees are yet to be done, and it's not clear yet how tree tokens could work. Also, the "offsets" or "span" for inputs without spaces have to be double checked

        self.tree_tokens = [Functive(segment,span,position,ids,semiotic) for segment,span,position,ids in zip(hf_output.tokens,hf_output.offsets,hf_output.word_ids,hf_output.ids)]

        self.tokens = sorted([token for token in self.tree_tokens if token.position != None], key= lambda x: x.position)

        chain = "".join(self.split_norm)
        tree_root = Functive(chain, (0, len(chain)), None, None, semiotic)
        tree_root.children = self.tokens

        for token in self.tokens:
            token.head = tree_root

        self.len = len(self.tokens)
        self.labels = [token.label for token in self.tokens]


        # self.segmented = " ".join(self.labels)

    @property
    def nodes_split(self):
        nodes_list = []
        for i in range(len(self.split_norm)):
            start_i = len("".join(self.split_norm[:i]))
            end_i = start_i + len(self.split_norm[i])
            nodes_list.append((
                self.split_norm[i],
                (start_i, end_i)
                ))
            
        return set(nodes_list)
    
    @property
    def nodes(self):
        return [(token.label, token.span) for token in self.tokens]
    
    def mask(self,n):
        """
        Outputs a string with the nth label(s) of the chain replaced with the mask token. n can be an integer or a list of integers.
        """
        if isinstance(n,int):
            n = [n]

        assert max(n)<self.len, f"SLG: The mask position ({max(n)}) is bigger than the length of the chain ({self.len})."

        mask_token = self.semiotic.config.vocabulary.mask_token

        masked_chain = " ".join([t if i not in n else mask_token for i,t in enumerate(self.labels)])
        
        return masked_chain   

    def mask_tokens(self,n):
        """
        Outputs a new list with the nth token(s) of the chain replaced with the mask token. n can be an integer or a list of integers.
        """
        if isinstance(n,int):
            n = [n]

        assert max(n)<self.len, f"SLG: The mask position ({max(n)}) is bigger than the length of the chain ({self.len})."

        mask_token = self.semiotic.config.vocabulary.mask_token
        # This id for mask_token depends on the tokenizer being able to include the mask_token, which, due to a HF bug, happens at the end (hence the high id). Double-check when the bug has been dealt with
        mask_token_id = self.semiotic.syntagmatic.tokenizer.token_to_id(mask_token)

        masked_chain = [token if i not in n else Functive(mask_token,token.span,token.position,mask_token_id,self.semiotic) for i,token in enumerate(self.tokens)]
        
        return masked_chain







        

    def __repr__(self) -> str:
        return f"Chain({self.input})"

    # def __str__(self) -> str:
    #     return self.segmented

    def __iter__(self):
       ''' Returns the Iterator object '''
       return ChainIterator(self)

    def __getitem__(self, index:str):
        return self.tokens[index]









class PostProcessor:
    """
    Base PostProcessor class
    """

    def __init__(self) -> None:
        pass

    def post_process(self, sequence):
        pass

class disable:
    """
    Disable this step. It returns the input as output #TODO: maybe with the correct type for the pipeline
    """

    def __init__(self) -> None:
        pass

    def post_process(self, sequence):
        return sequence

class WikiFR(PostProcessor):
    """
    Rules for cleaning the wikipedia corpus. Among others, only one occurrence of each sentence is kept, to avoid the statistical bias of titles and categories used by Wikipedia 

    Args:
        PostProcessor ([type]): [description]
    """
    def __init__(self) -> None:
        super().__init__()
    
    def post_process(self, sentences):

        return list({sent for sent in set(sentences)
                if
                all([
                not sent.startswith(("Catégorie:")),
                " " in sent,
                "│" not in sent,
                "|" not in sent,
                "─o" not in sent,
                ])
        })

class WikiEN(PostProcessor):
    """
    Rules for cleaning the wikipedia corpus. Among others, only one occurrence of each sentence is kept, to avoid the statistical bias of titles and categories used by Wikipedia 

    Args:
        PostProcessor ([type]): [description]
    """
    def __init__(self) -> None:
        super().__init__()
    
    def post_process(self, sentences):

        return list({sent for sent in set(sentences)
                if
                all([
                not sent.startswith(("Category:")),
                " " in sent,
                "│" not in sent,
                "|" not in sent,
                "─o" not in sent,
                ])
        })